from enum import Enum

import numba
from numba.extending import overload
import numba.core.types as nbtypes

from pfb.wavelets.common import NUMBA_SEQUENCE_TYPES


class Modes(Enum):
    zeropad = 0
    symmetric = 1
    constant_edge = 2
    smooth = 3
    periodic = 4
    periodisation = 4
    reflect = 5
    asymmetric = 6
    antireflect = 7


def mode_str_to_enum(mode_str):
    pass


@overload(mode_str_to_enum)
def mode_str_to_enum_impl(mode_str):
    if isinstance(mode_str, nbtypes.UnicodeType):
        # Modes.mode.name doesn't work in the jitted code
        # so expand it all.
        zeropad_name = "zero"  # pywt is zero not zeropad
        symmetric_name = Modes.symmetric.name
        constant_edge_name = Modes.constant_edge.name
        smooth_name = Modes.smooth.name
        periodic_name = Modes.periodic.name
        periodisation_name = Modes.periodisation.name
        reflect_name = Modes.reflect.name
        asymmetric_name = Modes.asymmetric.name
        antireflect_name = Modes.antireflect.name

        def impl(mode_str):
            mode_str = mode_str.lower()

            if mode_str == zeropad_name:
                return Modes.zeropad
            elif mode_str == symmetric_name:
                return Modes.symmetric
            elif mode_str == constant_edge_name:
                return Modes.constant_edge
            elif mode_str == smooth_name:
                return Modes.smooth
            elif mode_str == periodic_name:
                return Modes.periodic
            elif mode_str == periodisation_name:
                return Modes.periodisation
            elif mode_str == reflect_name:
                return Modes.reflect
            elif mode_str == asymmetric_name:
                return Modes.asymmetric
            elif mode_str == antireflect_name:
                return Modes.antireflect
            else:
                raise ValueError("Unknown mode string")

        return impl


@numba.generated_jit(nopython=True, nogil=True, cache=True)
def promote_mode(mode, naxis):
    if not isinstance(naxis, nbtypes.Integer):
        raise TypeError("naxis must be an integer")

    if isinstance(mode, nbtypes.misc.UnicodeType):
        def impl(mode, naxis):
            return numba.typed.List([mode_str_to_enum(mode)
                                     for _ in range(naxis)])

    elif isinstance(mode, nbtypes.misc.StringLiteral):
        literal_mode = mode.literal_value
        def impl(mode, naxis):
            return numba.typed.List([mode_str_to_enum(literal_mode)
                                     for _ in range(naxis)])

    elif (isinstance(mode, NUMBA_SEQUENCE_TYPES) and
            isinstance(mode.dtype, nbtypes.misc.UnicodeType)):

        def impl(mode, naxis):
            if len(mode) != naxis:
                raise ValueError("len(mode) != len(axis)")

            return numba.typed.List([mode_str_to_enum(m) for m in mode])

    else:

        raise TypeError("mode must be a string, "
                        "a list of strings "
                        "or a tuple of strings. "
                        "Got %s." % mode)

    return impl



import sys
import numpy as np
import numexpr as ne
from numba import jit, njit, prange
from numba.extending import overload
import dask
import dask.array as da
from dask.distributed import performance_report
from dask.diagnostics import ProgressBar
from ducc0.fft import r2c, c2r, good_size
iFs = np.fft.ifftshift
Fs = np.fft.fftshift
from daskms import xds_from_storage_ms as xds_from_ms
from daskms import xds_from_storage_table as xds_from_table
from daskms.experimental.zarr import xds_from_zarr
from omegaconf import ListConfig
from skimage.morphology import label
from scipy.optimize import curve_fit
from collections import namedtuple
from africanus.coordinates.coordinates import radec_to_lmn
import xarray as xr
from quartical.utils.dask import Blocker
from scipy.interpolate import RegularGridInterpolator
import sympy as sm
from sympy.utilities.lambdify import lambdify
from sympy.parsing.sympy_parser import parse_expr

JIT_OPTIONS = {
    "nogil": True,
    "fastmath": True,
    "cache": True
}

import pdb
class ForkedPdb(pdb.Pdb):
    """A Pdb subclass that may be used
    from a forked multiprocessing child

    """
    def interaction(self, *args, **kwargs):
        _stdin = sys.stdin
        try:
            sys.stdin = open('/dev/stdin')
            pdb.Pdb.interaction(self, *args, **kwargs)
        finally:
            sys.stdin = _stdin


def compute_context(scheduler, output_filename, boring=True):
    if scheduler == "distributed":
        return performance_report(filename=output_filename + "_dask_report.html")
    else:
        if boring:
            from contextlib import nullcontext
            return nullcontext()
        else:
            return ProgressBar()


def kron_matvec(A, b):
    D = len(A)
    N = b.size
    x = b

    for d in range(D):
        Gd = A[d].shape[0]
        NGd = N // Gd
        X = np.reshape(x, (Gd, NGd))
        Z = A[d].dot(X).T
        x = Z.ravel()
    return x.reshape(b.shape)

@jit(nopython=True, fastmath=True, parallel=False, cache=True, nogil=True)
def kron_matvec2(A, b):
    D = len(A)
    N = b.size
    x = b

    for d in range(D):
        Gd = A[d].shape[0]
        NGd = N//Gd
        X = np.reshape(x, (Gd, NGd))
        Z = np.zeros((Gd, NGd), dtype=A[0].dtype)
        Ad = A[d]
        for i in range(Gd):
            for j in range(Gd):
                for k in range(NGd):
                    Z[j, k] += Ad[i, j] * X[i, k]
        x[:] = Z.T.ravel()
    return x


def to4d(data):
    if data.ndim == 4:
        return data
    elif data.ndim == 2:
        return data[None, None]
    elif data.ndim == 3:
        return data[None]
    elif data.ndim == 1:
        return data[None, None, None]
    else:
        raise ValueError("Only arrays with ndim <= 4 can be broadcast to 4D.")


def Gaussian2D(xin, yin, GaussPar=(1., 1., 0.), normalise=True, nsigma=5):
    S0, S1, PA = GaussPar
    Smaj = S0  #np.maximum(S0, S1)
    Smin = S1  #np.minimum(S0, S1)
    # print(f'using ex = {Smaj}, ey = {Smin}')
    A = np.array([[1. / Smin ** 2, 0],
                  [0, 1. / Smaj ** 2]])

    c, s, t = np.cos, np.sin, np.deg2rad(-PA)
    R = np.array([[c(t), -s(t)],
                  [s(t), c(t)]])
    A = np.dot(np.dot(R.T, A), R)
    sOut = xin.shape
    # only compute the result out to 5 * emaj
    extent = (nsigma * Smaj)**2
    xflat = xin.squeeze()
    yflat = yin.squeeze()
    idx, idy = np.where(xflat**2 + yflat**2 <= extent)
    x = np.array([xflat[idx, idy].ravel(), yflat[idx, idy].ravel()])
    R = np.einsum('nb,bc,cn->n', x.T, A, x)
    # need to adjust for the fact that GaussPar corresponds to FWHM
    fwhm_conv = 2 * np.sqrt(2 * np.log(2))
    tmp = np.exp(-fwhm_conv * R)
    gausskern = np.zeros(xflat.shape, dtype=np.float64)
    gausskern[idx, idy] = tmp

    if normalise:
        gausskern /= np.sum(gausskern)
    return np.ascontiguousarray(gausskern.reshape(sOut),
                                dtype=np.float64)


def give_edges(p, q, nx, ny, nx_psf, ny_psf):
    nx0 = nx_psf//2
    ny0 = ny_psf//2

    # image overlap edges
    # left edge for x coordinate
    dxl = p - nx0
    xl = np.maximum(dxl, 0)

    # right edge for x coordinate
    dxu = p + nx0
    xu = np.minimum(dxu, nx)
    # left edge for y coordinate
    dyl = q - ny0
    yl = np.maximum(dyl, 0)
    # right edge for y coordinate
    dyu = q + ny0
    yu = np.minimum(dyu, ny)

    # PSF overlap edges
    xlpsf = np.maximum(nx0 - p , 0)
    xupsf = np.minimum(nx0 + nx - p, nx_psf)
    ylpsf = np.maximum(ny0 - q, 0)
    yupsf = np.minimum(ny0 + ny - q, ny_psf)

    return slice(xl, xu), slice(yl, yu), \
        slice(xlpsf, xupsf), slice(ylpsf, yupsf)


def get_padding_info(nx, ny, pfrac):
    npad_x = int(pfrac * nx)
    nfft = good_size(nx + npad_x, True)
    npad_xl = (nfft - nx) // 2
    npad_xr = nfft - nx - npad_xl

    npad_y = int(pfrac * ny)
    nfft = good_size(ny + npad_y, True)
    npad_yl = (nfft - ny) // 2
    npad_yr = nfft - ny - npad_yl
    padding = ((0, 0), (npad_xl, npad_xr), (npad_yl, npad_yr))
    unpad_x = slice(npad_xl, -npad_xr)
    unpad_y = slice(npad_yl, -npad_yr)
    return padding, unpad_x, unpad_y


def convolve2gaussres(image, xx, yy, gaussparf, nthreads, gausspari=None,
                      pfrac=0.5, norm_kernel=False):
    """
    Convolves the image to a specified resolution.

    Parameters
    ----------
    Image       - (nband, nx, ny) array to convolve
    xx/yy       - coordinates on the grid in the same units as gaussparf.
    gaussparf   - tuple containing Gaussian parameters of desired resolution
                  (emaj, emin, pa).
    gausspari   - initial resolution . By default it is assumed that the image
                  is a clean component image with no associated resolution.
                  If beampari is specified, it must be a tuple containing
                  gausspars for each imaging band in the same format.
    nthreads    - number of threads to use for the FFT's.
    pfrac       - padding used for the FFT based convolution.
                  Will pad by pfrac/2 on both sides of image
    """
    nband, nx, ny = image.shape
    padding, unpad_x, unpad_y = get_padding_info(nx, ny, pfrac)
    ax = (1, 2)  # axes over which to perform fft
    lastsize = ny + np.sum(padding[-1])

    gausskern = Gaussian2D(xx, yy, gaussparf, normalise=norm_kernel)
    gausskern = np.pad(gausskern[None], padding, mode='constant')
    gausskernhat = r2c(iFs(gausskern, axes=ax), axes=ax, forward=True,
                       nthreads=nthreads, inorm=0)

    image = np.pad(image, padding, mode='constant')
    imhat = r2c(iFs(image, axes=ax), axes=ax, forward=True, nthreads=nthreads,
                inorm=0)

    # convolve to desired resolution
    if gausspari is None:
        imhat *= gausskernhat
    else:
        for i in range(nband):
            thiskern = Gaussian2D(xx, yy, gausspari[i], normalise=norm_kernel)
            thiskern = np.pad(thiskern[None], padding, mode='constant')
            thiskernhat = r2c(iFs(thiskern, axes=ax), axes=ax, forward=True,
                              nthreads=nthreads, inorm=0)

            convkernhat = np.zeros_like(thiskernhat)
            msk = np.abs(thiskernhat) > 0.0
            convkernhat[msk] = gausskernhat[msk]/thiskernhat[msk]

            imhat[i] *= convkernhat[0]

    image = Fs(c2r(imhat, axes=ax, forward=False, lastsize=lastsize, inorm=2,
                   nthreads=nthreads), axes=ax)[:, unpad_x, unpad_y]

    return image


@dask.delayed
def fetch_poltype(corr_type):
    corr_type = set(tuple(corr_type))
    if corr_type.issubset(set([9, 10, 11, 12])):
        return 'linear'
    elif corr_type.issubset(set([5, 6, 7, 8])):
        return 'circular'


def construct_mappings(ms_name,
                       gain_name=None,
                       ipi=None,
                       cpi=None,
                       freq_min=-np.inf,
                       freq_max=np.inf):
    '''
    Construct dictionaries containing per MS, FIELD, DDID and SCAN
    time and frequency mappings.

    Input:
    ms_name     - list of ms names
    gain_name   - list of paths to gains, must be in same order as ms_names
    ipi         - integrations (i.e. unique times) per output image.
                  Defaults to one per scan.
    cpi         - Channels per image. Defaults to one per spw.

    The chan <-> band mapping is determined by:

    freqs           - dict[MS][IDT] frequencies chunked by band
    fbin_idx        - dict[MS][IDT] freq bin starting indices
    fbin_counts     - dict[MS][IDT] freq bin counts
    band_mapping    - dict[MS][IDT] output bands in dataset
    freq_out        - array of linearly spaced output frequencies
                      over entire frequency range.
                      freq_out[band_mapping[MS][IDT]] gives the
                      model frequencies in a dataset

    where IDT is constructed as FIELD#_DDID#_SCAN#.

    Similarly, the row <-> time mapping is determined by:

    utimes          - dict[MS][IDT] unique times per output image
    tbin_idx        - dict[MS][IDT] time bin starting indices
    tbin_counts     - dict[MS][IDT] time bin counts
    time_mapping    - dict[MS][IDT] utimes per dataset

    '''

    if not isinstance(ms_name, list) and not isinstance(ms_name, ListConfig):
        ms_name = [ms_name]

    if gain_name is not None:
        if not isinstance(gain_name, list) and not isinstance(gain_name, ListConfig):
            gain_name = [gain_name]
        assert len(ms_name) == len(gain_name)

    # collect times and frequencies per ms and ds
    freqs = {}
    chan_widths = {}
    times = {}
    gain_times = {}
    gain_freqs = {}
    gain_axes = {}
    gain_spec = {}
    radecs = {}
    antpos = {}
    poltype = {}
    uv_maxs = []
    idts = {}
    for ims, ms in enumerate(ms_name):
        xds = xds_from_ms(ms, chunks={"row": -1}, columns=('TIME','UVW'),
                          group_cols=['FIELD_ID', 'DATA_DESC_ID', 'SCAN_NUMBER'])

        # subtables
        ddids = xds_from_table(ms + "::DATA_DESCRIPTION")[0]
        fields = xds_from_table(ms + "::FIELD")[0]
        spws = xds_from_table(ms + "::SPECTRAL_WINDOW")[0]
        pols = xds_from_table(ms + "::POLARIZATION")[0]
        ants = xds_from_table(ms + "::ANTENNA")[0]

        antpos[ms] = ants.POSITION.data
        poltype[ms] = fetch_poltype(pols.CORR_TYPE.data.squeeze())

        idts[ms] = []
        if gain_name is not None:
            gain = xds_from_zarr(gain_name[ims])
            gain_times[ms] = {}
            gain_freqs[ms] = {}
            gain_axes[ms] = {}
            gain_spec[ms] = {}

        freqs[ms] = {}
        times[ms] = {}
        radecs[ms] = {}
        chan_widths[ms] = {}
        for ids, ds in enumerate(xds):
            idt = f"FIELD{ds.FIELD_ID}_DDID{ds.DATA_DESC_ID}_SCAN{ds.SCAN_NUMBER}"
            idts[ms].append(idt)
            radecs[ms][idt] = fields.PHASE_DIR.data[ds.FIELD_ID].squeeze()

            freqs[ms][idt] = spws.CHAN_FREQ.data[ds.DATA_DESC_ID]
            chan_widths[ms][idt] = spws.CHAN_WIDTH.data[ds.DATA_DESC_ID]
            times[ms][idt] = da.atleast_1d(ds.TIME.data.squeeze())
            uvw = ds.UVW.data
            u_max = abs(uvw[:, 0].max())
            v_max = abs(uvw[:, 1]).max()
            uv_maxs.append(da.maximum(u_max, v_max))

            if gain_name is not None:
                gain_times[ms][idt] = gain[ids].gain_time.data
                gain_freqs[ms][idt] = gain[ids].gain_freq.data
                gain_axes[ms][idt] = gain[ids].GAIN_AXES
                gain_spec[ms][idt] = gain[ids].GAIN_SPEC


    # Early compute to get metadata
    times, freqs, gain_times, gain_freqs, gain_axes, gain_spec, radecs,\
        chan_widths, uv_maxs, antpos, poltype = dask.compute(times, freqs,\
        gain_times, gain_freqs, gain_axes, gain_spec, radecs, chan_widths,\
        uv_maxs, antpos, poltype)

    uv_max = max(uv_maxs)
    all_freqs = []
    all_times = []
    freq_mapping = {}
    row_mapping = {}
    time_mapping = {}
    utimes = {}
    ms_chunks = {}
    gain_chunks = {}
    for ms in ms_name:
        freq_mapping[ms] = {}
        row_mapping[ms] = {}
        time_mapping[ms] = {}
        utimes[ms] = {}
        ms_chunks[ms] = []
        gain_chunks[ms] = []
        for idt in idts[ms]:
            freq = freqs[ms][idt]
            if gain_name is not None:
                try:
                    assert (gain_freqs[ms][idt] == freq).all()
                except Exception as e:
                    raise ValueError(f'Mismatch between gain and MS '
                                     f'frequencies for {ms} at {idt}')

            nchan_in = freq.size
            idx = (freq>=freq_min) & (freq<=freq_max)
            if not idx.any():
                continue
            idx0 = np.argmax(idx) # returns index of first True element
            # np.searchsorted here?
            try:
                # returns zero if not idx.any()
                assert idx[idx0]
            except Exception as e:
                continue
            freq = freq[idx]
            nchan = freq.size
            if cpi in [-1, 0, None]:
                cpit = nchan
                cpi = nchan_in
            else:
                cpit = np.minimum(cpi, nchan)
                cpi = np.minimum(cpi, nchan_in)
            freq_mapping[ms][idt] = {}
            tmp = np.arange(idx0, idx0 + nchan, cpit)
            freq_mapping[ms][idt]['start_indices'] = tmp
            if cpit != nchan:
                tmp2 = np.append(tmp, [idx0 + nchan])
                freq_mapping[ms][idt]['counts'] = tmp2[1:] - tmp2[0:-1]
            else:
                freq_mapping[ms][idt]['counts'] = np.array((nchan,), dtype=int)

            time = times[ms][idt]
            utime = np.unique(time)
            if gain_name is not None:
                try:
                    assert (gain_times[ms][idt] == utime).all()
                except Exception as e:
                    raise ValueError(f'Mismatch between gain and MS '
                                     f'utimes for {ms} at {idt}')
            utimes[ms][idt] = utime
            all_times.append(utime)

            ntime = utimes[ms][idt].size
            if ipi in [0, -1, None]:
                ipit = ntime
            else:
                ipit = np.minimum(ipi, ntime)
            row_chunks, ridx, rcounts = chunkify_rows(time, ipit,
                                                      daskify_idx=False)
            # these are for applying gains
            # essentially the number of rows per unique time
            row_mapping[ms][idt] = {}
            row_mapping[ms][idt]['start_indices'] = ridx
            row_mapping[ms][idt]['counts'] = rcounts

            nfreq_chunks = nchan_in // cpi
            freq_chunks = (cpi,)*nfreq_chunks
            rem = nchan_in - nfreq_chunks * cpi
            if rem:
                freq_chunks += (rem,)

            ms_chunks[ms].append({'row': row_chunks,
                                  'chan': freq_chunks})

            time_mapping[ms][idt] = {}
            tmp = np.arange(0, ntime, ipit)
            time_mapping[ms][idt]['start_indices'] = tmp
            if ipit != ntime:
                tmp2 = np.append(tmp, [ntime])
                time_mapping[ms][idt]['counts'] = tmp2[1:] - tmp2[0:-1]
            else:
                time_mapping[ms][idt]['counts'] = np.array((ntime,))

            # we may need to rechunk gains in time and freq to line up with MS
            if gain_name is not None:
                tmp_dict = {}
                for name, val in zip(gain_axes[ms][idt], gain_spec[ms][idt]):
                    if name == 'gain_time':
                        tmp_dict[name] = tuple(time_mapping[ms][idt]['counts'])
                    elif name == 'gain_freq':
                        tmp_dict[name] = freq_chunks
                    elif name == 'direction':
                        if len(val) > 1:
                            raise ValueError("DD gains not supported yet")
                        if val[0] > 1:
                            raise ValueError("DD gains not supported yet")
                        tmp_dict[name] = tuple(val)
                    else:
                        tmp_dict[name] = tuple(val)

                gain_chunks[ms].append(tmp_dict)

    return row_mapping, freq_mapping, time_mapping, \
           freqs, utimes, ms_chunks, gain_chunks, radecs, \
           chan_widths, uv_max, antpos, poltype


def restore_corrs(vis, ncorr):
    return da.blockwise(_restore_corrs, ('row', 'chan', 'corr'),
                        vis, ('row', 'chan'),
                        ncorr, None,
                        new_axes={"corr": ncorr},
                        dtype=vis.dtype)


def _restore_corrs(vis, ncorr):
    model_vis = np.zeros(vis.shape+(ncorr,), dtype=vis.dtype)
    model_vis[:, :, 0] = vis
    if model_vis.shape[-1] > 1:
        model_vis[:, :, -1] = vis
    return model_vis

def fitcleanbeam(psf: np.ndarray,
                 level: float = 0.5,
                 pixsize: float = 1.0,
                 extent: float = 15.0):
    """
    Find the Gaussian that approximates the PSF.
    First find the main lobe by identifying where PSF > level
    then fit Gaussian out to a radius of extent * max(x, y) where
    x and y are the coordinates where PSF > level.
    """
    nband, nx, ny = psf.shape

    # pixel coordinates
    x = np.arange(-nx / 2, nx / 2)
    y = np.arange(-ny / 2, ny / 2)
    xx, yy = np.meshgrid(x, y, indexing='ij')

    # model to fit
    def func(xy, emaj, emin, pa):
        Smin = np.minimum(emaj, emin)
        Smaj = np.maximum(emaj, emin)

        A = np.array([[1. / Smin ** 2, 0],
                      [0, 1. / Smaj ** 2]])

        c, s, t = np.cos, np.sin, np.deg2rad(-pa)
        R = np.array([[c(t), -s(t)],
                      [s(t), c(t)]])
        A = np.dot(np.dot(R.T, A), R)
        R = np.einsum('nb,bc,cn->n', xy.T, A, xy)
        # GaussPar should corresponds to FWHM
        fwhm_conv = 2 * np.sqrt(2 * np.log(2))
        return np.exp(-fwhm_conv * R)

    Gausspars = []
    for v in range(nband):
        # make sure psf is normalised
        psfv = psf[v] / psf[v].max()
        # find regions where psf is above level
        mask = np.where(psfv > level, 1.0, 0)

        # label all islands and find center
        islands = label(mask)
        ncenter = islands[nx // 2, ny // 2]

        # get extend of main lobe
        x = xx[islands == ncenter]
        y = yy[islands == ncenter]
        xdiff = x.max() - x.min()
        ydiff = y.max() - y.min()
        rsq = np.abs(x).max()**2 + np.abs(y).max()**2
        rrsq = xx**2 + yy**2
        idxs = rrsq < extent * rsq

        # select psf main lobe
        psfv = psfv[idxs]
        x = xx[idxs]
        y = yy[idxs]
        xy = np.vstack((x, y))
        emaj0 = np.maximum(xdiff, ydiff)
        emin0 = np.minimum(xdiff, ydiff)
        p, _ = curve_fit(func, xy, psfv, p0=(emaj0, emin0, 0.0))
        Gausspars.append([p[0] * pixsize, p[1] * pixsize, p[2]])

    return Gausspars


# utility to produce a model image from fitted components
def model_from_comps(comps, freq, mask, band_mapping, ref_freq, fitted):
    '''
    comps - (order/nband, ncomps) array of fitted components or model vals
            per band if no fitting was done.

    freq - (nchan) array of output frequencies

    mask - (nx, ny) bool array indicating where model is non-zero.
           comps need to be aligned with this mask i.e.
           Ix, Iy = np.where(mask) are the xy locations of non-zero pixels
           model[:, Ix, Iy] = comps if no fitting was done

    band_mapping - tuple containg the indices of bands mapping to the output.
                   In the case of a single spectral window we usually have
                   band_mapping = np.arange(nband).

    ref_freq - reference frequency used during fitting i.e. we made the
                design matrix using freq/ref_freq as coordinate

    fitted - bool indicating if any fitting actually happened.
    '''
    return da.blockwise(_model_from_comps_wrapper, ('out', 'nx', 'ny'),
                        comps, ('com', 'pix'),
                        freq, ('chan',),
                        mask, ('nx', 'ny'),
                        band_mapping, ('out',),
                        ref_freq, None,
                        fitted, None,
                        align_arrays=False,
                        dtype=comps.dtype)


def _model_from_comps_wrapper(comps, freq, mask, band_mapping, ref_freq, fitted):
    return _model_from_comps(comps[0][0], freq[0], mask, band_mapping, ref_freq, fitted)


def _model_from_comps(comps, freq, mask, band_mapping, ref_freq, fitted):
    freqo = freq[band_mapping]
    nband = freqo.size
    nx, ny = mask.shape
    model = np.zeros((nband, nx, ny), dtype=comps.dtype)
    if fitted:
        order, npix = comps.shape
        nband = freqo.size
        nx, ny = mask.shape
        model = np.zeros((nband, nx, ny), dtype=comps.dtype)
        w = (freqo / ref_freq).reshape(nband, 1)
        Xdes = np.tile(w, order) ** np.arange(0, order)
        beta_rec = Xdes.dot(comps)
        model[:, mask] = beta_rec
    else:
        model[:, mask] = comps[band_mapping]

    return model


def init_mask(mask, model, output_type, log):
    if mask is None:
        print("No mask provided", file=log)
        mask = np.ones((mds.nx, mds.ny), dtype=output_type)
    elif mask.endswith('.fits'):
        try:
            mask = load_fits(mask, dtype=output_type).squeeze()
            assert mask.shape == (mds.nx, mds.ny)
            print('Using provided fits mask', file=log)
        except Exception as e:
            print(f"No mask found at {mask}", file=log)
            raise e
    elif mask.lower() == 'model':
        mask = np.any(model, axis=0)
        print('Using model to construct mask', file=log)
    else:
        raise ValueError(f'Unsupported masking option {mask}')
    return mask


def dds2cubes(dds, nband, apparent=False):
    real_type = dds[0].DIRTY.dtype
    complex_type = np.result_type(real_type, np.complex64)
    nx, ny = dds[0].DIRTY.shape
    dirty = [da.zeros((nx, ny), chunks=(-1, -1),
                      dtype=real_type) for _ in range(nband)]
    model = [da.zeros((nx, ny), chunks=(-1, -1),
                      dtype=real_type) for _ in range(nband)]
    if 'RESIDUAL' in dds[0]:
        residual = [da.zeros((nx, ny), chunks=(-1, -1),
                            dtype=real_type) for _ in range(nband)]
    else:
        residual = None
    wsums = [da.zeros(1, dtype=real_type) for _ in range(nband)]
    if 'PSF' in dds[0]:
        nx_psf, ny_psf = dds[0].PSF.shape
        nx_psf, nyo2_psf = dds[0].PSFHAT.shape
        psf = [da.zeros((nx_psf, ny_psf), chunks=(-1, -1),
                        dtype=real_type) for _ in range(nband)]
        psfhat = [da.zeros((nx_psf, nyo2_psf), chunks=(-1, -1),
                            dtype=complex_type) for _ in range(nband)]
    else:
        psf = None
        psfhat = None
    mean_beam = [da.zeros((nx, ny), chunks=(-1, -1),
                            dtype=real_type) for _ in range(nband)]
    if 'DUAL' in dds[0]:
        nbasis, nmax = dds[0].DUAL.shape
        dual = [da.zeros((nbasis, nmax), chunks=(-1, -1),
                            dtype=real_type) for _ in range(nband)]
    else:
        dual = None
    for ds in dds:
        b = ds.bandid
        if apparent:
            dirty[b] += ds.DIRTY.data
            if 'RESIDUAL' in ds:
                residual[b] += ds.RESIDUAL.data
        else:
            dirty[b] += ds.DIRTY.data * ds.BEAM.data
            if 'RESIDUAL' in ds:
                residual[b] += ds.RESIDUAL.data * ds.BEAM.data
        if 'PSF' in ds:
            psf[b] += ds.PSF.data
            psfhat[b] += ds.PSFHAT.data
        if 'MODEL' in ds:
            model[b] = ds.MODEL.data
        if 'DUAL' in ds:
            dual[b] = ds.DUAL.data
        mean_beam[b] += ds.BEAM.data * ds.WSUM.data[0]
        wsums[b] += ds.WSUM.data[0]
    wsums = da.stack(wsums).reshape(nband)
    wsum = wsums.sum()
    dirty = da.stack(dirty)/wsum
    model = da.stack(model)
    if 'RESIDUAL' in ds:
        residual = da.stack(residual)/wsum
    if 'PSF' in ds:
        psf = da.stack(psf)/wsum
        psfhat = da.stack(psfhat)/wsum
    if 'DUAL' in ds:
        dual = da.stack(dual)
    for b in range(nband):
        if wsums[b]:
            mean_beam[b] /= wsums[b]
    mean_beam = da.stack(mean_beam)
    dirty, model, residual, psf, psfhat, mean_beam, wsums, dual = dask.compute(
                                                                dirty,
                                                                model,
                                                                residual,
                                                                psf,
                                                                psfhat,
                                                                mean_beam,
                                                                wsums,
                                                                dual)
    return dirty, model, residual, psf, psfhat, mean_beam, wsums, dual


def chunkify_rows(time, utimes_per_chunk, daskify_idx=False):
    utimes, time_bin_counts = np.unique(time, return_counts=True)
    n_time = len(utimes)
    if utimes_per_chunk == 0 or utimes_per_chunk == -1:
        utimes_per_chunk = n_time
    row_chunks = [np.sum(time_bin_counts[i:i+utimes_per_chunk])
                  for i in range(0, n_time, utimes_per_chunk)]
    time_bin_indices = np.zeros(n_time, dtype=np.int32)
    time_bin_indices[1::] = np.cumsum(time_bin_counts)[0:-1]
    time_bin_indices = time_bin_indices.astype(np.int32)
    time_bin_counts = time_bin_counts.astype(np.int32)
    if daskify_idx:
        time_bin_indices = da.from_array(time_bin_indices, chunks=utimes_per_chunk)
        time_bin_counts = da.from_array(time_bin_counts, chunks=utimes_per_chunk)
    return tuple(row_chunks), time_bin_indices, time_bin_counts


def rephase_vis(vis, uvw, radec_in, radec_out):
    return da.blockwise(_rephase_vis, 'rf',
                        vis, 'rf',
                        uvw, 'r3',
                        radec_in, None,
                        radec_out, None,
                        dtype=vis.dtype)

def _rephase_vis(vis, uvw, radec_in, radec_out):
    l_in, m_in, n_in = radec_to_lmn(radec_in)
    l_out, m_out, n_out = radec_to_lmn(radec_out)
    return vis * np.exp(1j*(uvw[:, 0]*(l_out-l_in) +
                            uvw[:, 1]*(m_out-m_in) +
                            uvw[:, 2]*(n_out-n_in)))


# TODO - should allow coarsening to values other than 1
def concat_row(xds):
    # TODO - how to compute average beam before we have access to grid?
    times_in = []
    freqs = []
    for ds in xds:
        times_in.append(ds.time_out)
        freqs.append(ds.freq_out)

    times_in = np.unique(times_in)
    freqs = np.unique(freqs)

    nband = freqs.size
    ntime_in = times_in.size

    if ntime_in == 1:  # no need to concatenate
        return xds

    # do merge manually because different variables require different
    # treatment anyway eg. the BEAM should be computed as a weighted sum
    xds_out = []
    for b in range(nband):
        xdsb = []
        times = []
        nu = freqs[b]
        for ds in xds:
            if ds.freq_out == nu:
                xdsb.append(ds)
                times.append(ds.time_out)

        wgt = [ds.WEIGHT for ds in xdsb]
        vis = [ds.VIS for ds in xdsb]
        mask = [ds.MASK for ds in xdsb]
        uvw = [ds.UVW for ds in xdsb]

        wgto = xr.concat(wgt, dim='row')
        viso = xr.concat(vis, dim='row')
        masko = xr.concat(mask, dim='row')
        uvwo = xr.concat(uvw, dim='row')

        xdso = xr.merge((wgto, viso, masko, uvwo))
        xdso['BEAM'] = xdsb[0].BEAM  # we need the grid to do this properly
        xdso['FREQ'] = xdsb[0].FREQ  # is this always going to be the case?

        xdso = xdso.chunk({'row':-1})

        xdso = xdso.assign_coords({
            'chan': (('chan',), xdsb[0].chan.data)
        })

        times = np.array(times)
        tout = np.round(np.mean(times), 5)  # avoid precision issues
        xdso = xdso.assign_attrs({
            'dec': xdsb[0].dec,  # always the case?
            'ra': xdsb[0].ra,    # always the case?
            'time_out': tout,
            'time_max': times.max(),
            'time_min': times.min(),
            'timeid': 0,
            'freq_out': nu,
            'freq_max': xdsb[0].freq_max,
            'freq_min': xdsb[0].freq_min,
        })
        xds_out.append(xdso)
    return xds_out


def concat_chan(xds, nband_out=1):
    times = []
    freqs_in = []
    freqs_min = []
    freqs_max = []
    all_freqs = []
    for ds in xds:
        times.append(ds.time_out)
        freqs_in.append(ds.freq_out)
        freqs_min.append(ds.freq_min)
        freqs_max.append(ds.freq_max)
        all_freqs.append(ds.chan)

    times = np.unique(times)
    freqs_in = np.unique(freqs_in)
    freqs_min = np.unique(freqs_min)
    freqs_max = np.unique(freqs_max)
    all_freqs = np.unique(all_freqs)

    nband_in = freqs_in.size
    ntime = times.size

    if nband_in == nband_out or nband_in == 1:  # no need to concatenate
        return xds

    # currently assuming linearly spaced frequencies
    freq_bins = np.linspace(freqs_min.min(), freqs_max.max(), nband_out+1)
    bin_centers = (freq_bins[1:] + freq_bins[0:-1])/2

    xds_out = []
    for t in range(ntime):
        time = times[t]
        for b in range(nband_out):
            xdst = []
            flow = freq_bins[b]
            fhigh = freq_bins[b+1]
            freqsb = all_freqs[all_freqs >= flow]
            # exlusive except for the last one
            if b==nband_out-1:
                freqsb = freqsb[freqsb <= fhigh]
            else:
                freqsb = freqsb[freqsb < fhigh]
            for ds in xds:
                # ds overlaps output if either ds.freq_min or ds.freq_max lies in the bin
                low_in = ds.freq_min > flow and ds.freq_min < fhigh
                high_in = ds.freq_max > flow and ds.freq_max < fhigh
                if ds.time_out == time and (low_in or high_in):
                    xdst.append(ds)

            # LB - we should be able to avoid this stack operation by using Jon's *() magic
            wgt = da.stack([ds.WEIGHT.data for ds in xdst]).rechunk(-1, -1) # verify chunking over freq axis
            vis = da.stack([ds.VIS.data for ds in xdst]).rechunk(-1, -1)
            mask = da.stack([ds.MASK.data for ds in xdst]).rechunk(-1, -1)
            freq = da.stack([ds.FREQ.data for ds in xdst]).rechunk(-1)

            nrow = xdst[0].row.size
            nchan = freqsb.size

            freqs_dask = da.from_array(freqsb, chunks=nchan)
            blocker = Blocker(sum_overlap, 's')
            blocker.add_input('vis', vis, 'src')
            blocker.add_input('wgt', wgt, 'src')
            blocker.add_input('mask', mask, 'src')
            blocker.add_input('freq', freq, 'sc')
            blocker.add_input('ufreq', freqs_dask, 'f')
            blocker.add_input('flow', flow, None)
            blocker.add_input('fhigh', fhigh, None)
            blocker.add_output('viso', 'rf', ((nrow,), (nchan,)), vis.dtype)
            blocker.add_output('wgto', 'rf', ((nrow,), (nchan,)), wgt.dtype)
            blocker.add_output('masko', 'rf', ((nrow,), (nchan,)), mask.dtype)

            out_dict = blocker.get_dask_outputs()

            data_vars = {
                'VIS': (('row', 'chan'), out_dict['viso']),
                'WEIGHT': (('row', 'chan'), out_dict['wgto']),
                'MASK': (('row', 'chan'), out_dict['masko']),
                'FREQ': (('chan',), freqs_dask),
                'UVW': (('row', 'three'), xdst[0].UVW.data), # should be the same across data sets
                'BEAM': (('scalar',), xdst[0].BEAM.data)  # need to pass in the grid to do this properly
            }

            coords = {
                'chan': (('chan',), freqsb)
            }

            fout = np.round(bin_centers[b], 5)  # avoid precision issues
            attrs = {
                'freq_out': fout,
                'bandid': b,
                'dec': xdst[0].dec,
                'ra': xdst[0].ra,
                'time_out': time
            }

            xdso = xr.Dataset(data_vars=data_vars,
                              attrs=attrs)

            xds_out.append(xdso)
    return xds_out


def sum_overlap(vis, wgt, mask, freq, ufreq, flow, fhigh):
    nds = vis.shape[0]

    # output grids
    nchan = ufreq.size
    nrow = vis.shape[1]
    viso = np.zeros((nrow, nchan), dtype=vis.dtype)
    wgto = np.zeros((nrow, nchan), dtype=wgt.dtype)
    masko = np.zeros((nrow, nchan), dtype=mask.dtype)

    # weighted sum at overlap
    for i in range(nds):
        nu = freq[i]
        _, idx0, idx1 = np.intersect1d(nu, ufreq, assume_unique=True, return_indices=True)
        viso[:, idx1] += vis[i][:, idx0] * wgt[i][:, idx0] * mask[i][:, idx0]
        wgto[:, idx1] += wgt[i][:, idx0] * mask[i][:, idx0]
        masko[:, idx1] += mask[i][:, idx0]

    # unmasked where at least one data point is unflagged
    masko = np.where(masko > 0, True, False)
    viso[masko] = viso[masko]/wgto[masko]

    # blocker expects a dictionary as output
    out_dict = {}
    out_dict['viso'] = viso
    out_dict['wgto'] = wgto
    out_dict['masko'] = masko.astype(np.uint8)

    return out_dict


def l1reweight_func(psiH, outvar, rmsfactor, rms_comps, model):
    '''
    The logic here is that weights should remain the same for model
    components that are rmsfactor times larger than the rms.
    High SNR values should experience relatively small thresholding
    whereas small values should be strongly thresholded
    '''
    psiH(model, outvar)
    mcomps = np.sum(outvar, axis=0)
    # the **2 here results in more agressive reweighting
    return (1 + rmsfactor)/(1 + mcomps**4/rms_comps**4)


# TODO - this can be done in parallel by splitting the image into facets
def fit_image_cube(time, freq, image, wgt=None, nbasist=None, nbasisf=None,
                   method='poly', sigmasq=0):
    '''
    Fit the time and frequency axes of an image cube where

    image   - (ntime, nband, nx, ny) pixelated image
    wgt     - (ntime, nband) optional per time and frequency weights
    nbasist - number of time basis functions
    nbasisf - number of frequency basis functions
    method  - method to use for fitting

    If wgt is not supplied equal weights are assumed.
    If nbasist/f are not supplied we return the coefficients
    of a bilinear fit regardless of method.

    methods:
    poly    - fit a monomials in time and frequency


    returns:
    coeffs  - fitted coefficients
    locx    - x pixel values
    locy    - y pixel values
    expr    - a string representing the symbolic expression describing the fit
    params  - tuple of str, parameters to pass into function (excluding t and f)
    ref_time    - reference time
    ref_freq    - reference frequency


    The fit is performed in scaled coordinates (t=time/ref_time,f=freq/ref_freq)
    '''
    ntime = time.size
    nband = freq.size
    ref_time = time[0]
    ref_freq = freq[0]
    import sympy as sm
    from sympy.abc import a, t, f

    if nbasist is None:
        nbasist = ntime
    else:
        assert nbasist <= ntime
    if nbasisf is None:
        nbasisf = nband
    else:
        assert nbasisf <= nband

    mask = np.any(image, axis=(0,1))  # over t and f axes
    Ix, Iy = np.where(mask)
    ncomps = Ix.size

    # components excluding zeros
    beta = image[:, :, Ix, Iy].reshape(ntime*nband, ncomps)
    if wgt is not None:
        wgt = wgt.reshape(ntime*nband, 1)
    else:
        wgt = np.ones((ntime*nband, 1), dtype=float)
    # nothing to fit
    if ntime==1 and nband==1:
        coeffs = beta
        expr = a
        params = (a,)
    elif method=='poly':
        wt = time/ref_time
        tfunc = t/ref_time
        Xfit = np.tile(wt[:, None], (nband, nbasist))**np.arange(nbasist)
        params = sm.symbols(f't(0:{nbasist})')
        expr = sum(co*t**i for i, co in enumerate(params))
        # the costant offset will always be included since nbasist is at least one
        if nband > 1:
            wf = freq/ref_freq
            ffunc = f/ref_freq
            Xf = np.tile(wf[:, None], (ntime, nbasisf-1))**np.arange(1, nbasisf)
            Xfit = np.hstack((Xfit, Xf))
            paramsf = sm.symbols(f'f(1:{nbasisf})')
            expr += sum(co*f**(i+1) for i, co in enumerate(paramsf))
            params += paramsf

    elif method=='Legendre':
        # scale to lie between -1,1 for stability
        if ntime > 1:
            tmax = time.max()
            tmin = time.min()
            wt = (time - (tmax + tmin)/2)
            wtmax = wt.max()
            wt /= wtmax
            # function to convert time to interp domain
            tfunc = (t - (tmax + tmin)/2)/wtmax
        else:
            wt = time
            tfunc = t
        Xt = np.zeros((ntime, nbasist), dtype=float)
        params = sm.symbols(f't(0:{nbasist})')
        if nbasist > 1:
            expr = 0
            for i in range(nbasist):
                vals = np.polynomial.Legendre.basis(i)(wt)
                Xt[:, i] = vals
                expr += sm.polys.orthopolys.legendre_poly(i, t)*params[i]
        else:
            Xt[...] = 1.0
            expr = params[0]
        Xfit = np.tile(Xt, (nband, 1))
        paramsf = sm.symbols(f'f(1:{nbasisf})')
        if nband > 1:
            Xf = np.zeros((nband, nbasisf - 1))
            fmax = freq.max()
            fmin = freq.min()
            wf = freq - (fmax + fmin)/2
            wfmax = wf.max()
            wf /= wfmax
            ffunc = (f - (fmax + fmin)/2)/wfmax
            for i in range(1, nbasisf):
                vals = np.polynomial.Legendre.basis(i)(wf)
                Xf[:, i-1] = vals
                expr += sm.polys.orthopolys.legendre_poly(i, f)*paramsf[i-1]
            Xf = np.tile(Xf, (ntime, 1))
            Xfit = np.hstack((Xfit, Xf))
            params += paramsf
    else:
        raise NotImplementedError("Please help us!")

    dirty_coeffs = Xfit.T.dot(wgt*beta)
    hess_coeffs = Xfit.T.dot(wgt*Xfit)
    # to improve conditioning
    if sigmasq:
        hess_coeffs += sigmasq*np.eye(hess_coeffs.shape[0])
    coeffs = np.linalg.solve(hess_coeffs, dirty_coeffs)


    return coeffs, Ix, Iy, str(expr), list(map(str,params)), str(tfunc),str(ffunc)


def eval_coeffs_to_cube(time, freq, nx, ny, coeffs, Ix, Iy,
                        expr, paramf, texpr, fexpr):
    ntime = time.size
    nfreq = freq.size

    image = np.zeros((ntime, nfreq, nx, ny), dtype=float)
    params = sm.symbols(('t','f'))
    params += sm.symbols(tuple(paramf))
    symexpr = parse_expr(expr)
    modelf = lambdify(params, symexpr)
    texpr = parse_expr(texpr)
    tfunc = lambdify(params[0], texpr)
    fexpr = parse_expr(fexpr)
    ffunc = lambdify(params[1], fexpr)
    for i, tval in enumerate(time):
        for j, fval in enumerate(freq):
            image[i, j, Ix, Iy] = modelf(tfunc(tval), ffunc(fval), *coeffs)

    return image


def eval_coeffs_to_slice(time, freq, coeffs, Ix, Iy,
                         expr, paramf, texpr, fexpr,
                         nxi, nyi, cellxi, cellyi, x0i, y0i,
                         nxo, nyo, cellxo, cellyo, x0o, y0o):

    image_in = np.zeros((nxi, nyi), dtype=float)
    params = sm.symbols(('t','f'))
    params += sm.symbols(tuple(paramf))
    symexpr = parse_expr(expr)
    modelf = lambdify(params, symexpr)
    texpr = parse_expr(texpr)
    tfunc = lambdify(params[0], texpr)
    fexpr = parse_expr(fexpr)
    ffunc = lambdify(params[1], fexpr)
    image_in[Ix, Iy] = modelf(tfunc(time), ffunc(freq), *coeffs)

    xin = (-(nxi//2) + np.arange(nxi))*cellxi + x0i
    yin = (-(nyi//2) + np.arange(nyi))*cellyi + y0i
    xo = (-(nxo//2) + np.arange(nxo))*cellxo + x0o
    yo = (-(nyo//2) + np.arange(nyo))*cellyo + y0o

    # how many pixels to pad by to extrapolate with zeros
    xldiff = xin.min() - xo.min()
    if xldiff > 0.0:
        npadxl = int(np.ceil(xldiff/cellxi))
    else:
        npadxl = 0
    yldiff = yin.min() - yo.min()
    if yldiff > 0.0:
        npadyl = int(np.ceil(yldiff/cellyi))
    else:
        npadyl = 0

    xudiff = xo.max() - xin.max()
    if xudiff > 0.0:
        npadxu = int(np.ceil(xudiff/cellxi))
    else:
        npadxu = 0
    yudiff = yo.max() - yin.max()
    if yudiff > 0.0:
        npadyu = int(np.ceil(yudiff/cellyi))
    else:
        npadyu = 0

    do_pad = npadxl > 0
    do_pad |= npadxu > 0
    do_pad |= npadyl > 0
    do_pad |= npadyu > 0
    if do_pad:
        image_in = np.pad(image_in,
                        ((npadxl, npadxu), (npadyl, npadyu)),
                        mode='constant')

        xin = (-(nxi//2+npadxl) + np.arange(nxi + npadxl + npadxu))*cellxi + x0i
        nxi = nxi + npadxl + npadxu
        yin = (-(nyi//2+npadyl) + np.arange(nyi + npadyl + npadyu))*cellyi + y0i
        nyi = nyi + npadyl + npadyu

    do_interp = cellxi != cellxo
    do_interp |= cellyi != cellyo
    do_interp |= x0i != x0o
    do_interp |= y0i != y0o
    do_interp |= nxi != nxo
    do_interp |= nyi != nyo
    if do_interp:
        interpo = RegularGridInterpolator((xin, yin), image_in,
                                          bounds_error=True, method='linear')
        xx, yy = np.meshgrid(xo, yo, indexing='ij')
        return interpo((xx, yy))
    # elif (nxi != nxo) or (nyi != nyo):
    #     # only need the overlap in this case
    #     _, idx0, idx1 = np.intersect1d(xin, xo, assume_unique=True, return_indices=True)
    #     _, idy0, idy1 = np.intersect1d(yin, yo, assume_unique=True, return_indices=True)
    #     return image[idx0, idy0]
    else:
        return image_in


@njit(**JIT_OPTIONS)
def norm_diff(x, xp):
    return norm_diff_impl(x, xp)


def norm_diff_impl(x, xp):
    return NotImplementedError


@overload(norm_diff_impl, jit_options=JIT_OPTIONS)
def nb_norm_diff_impl(x, xp):
    if x.ndim==3:
        def impl(x, xp):
            nband, nx, ny = x.shape
            num = 0.0
            den = 0.0
            for b in range(nband):
                for i in range(nx):
                    for j in range(ny):
                        num += (x[b, i, j] - xp[b, i, j])**2
                        den += x[b, i, j]**2
            return np.sqrt(num/den)
    elif x.ndim==2:
        def impl(x, xp):
            nx, ny = x.shape
            num = 0.0
            den = 0.0
            for i in range(nx):
                for j in range(ny):
                    num += (x[i, j] - xp[i, j])**2
                    den += x[i, j]**2
            return np.sqrt(num/den)
    else:
        raise ValueError("norm_diff is only implemented for 2D or 3D arrays")

    return impl



def remove_large_islands(x, max_island_size=100):
    islands = label(x.squeeze())
    num_islands = islands.max()
    for i in range(1,num_islands+1):
        msk = islands == i
        num_pix = np.sum(msk)
        if num_pix > max_island_size:
            x[msk] = 0.0
    return x



import pytest
from pathlib import Path
from xarray import Dataset
from collections import namedtuple
import dask
import dask.array as da
from daskms.experimental.zarr import xds_to_zarr, xds_from_zarr
pmp = pytest.mark.parametrize

@pmp('do_gains', (True, False))
def test_clean(do_gains, ms_name):
    '''
    Here we test that clean correctly infers the fluxes of point sources
    placed at the centers of pixels in the presence of the wterm and DI gain
    corruptions.
    TODO - add per scan PB variations
    '''

    import numpy as np
    np.random.seed(420)
    from numpy.testing import assert_allclose
    from daskms import xds_from_ms, xds_from_table, xds_to_table
    from daskms.experimental.zarr import xds_to_zarr
    from africanus.constants import c as lightspeed


    test_dir = Path(ms_name).resolve().parent
    xds = xds_from_ms(ms_name,
                      chunks={'row': -1, 'chan': -1})[0]
    spw = xds_from_table(f'{ms_name}::SPECTRAL_WINDOW')[0]

    utime = np.unique(xds.TIME.values)
    freq = spw.CHAN_FREQ.values.squeeze()
    freq0 = np.mean(freq)

    ntime = utime.size
    nchan = freq.size
    nant = np.maximum(xds.ANTENNA1.values.max(), xds.ANTENNA2.values.max()) + 1

    ncorr = xds.corr.size

    uvw = xds.UVW.values
    nrow = uvw.shape[0]
    u_max = abs(uvw[:, 0]).max()
    v_max = abs(uvw[:, 1]).max()
    uv_max = np.maximum(u_max, v_max)

    # image size
    cell_N = 1.0 / (2 * uv_max * freq.max() / lightspeed)

    srf = 2.0
    cell_rad = cell_N / srf
    cell_deg = cell_rad * 180 / np.pi
    cell_size = cell_deg * 3600
    print("Cell size set to %5.5e arcseconds" % cell_size)

    from ducc0.fft import good_size
    # the test will fail in intrinsic if sources fall near beam sidelobes
    fov = 1.0
    npix = good_size(int(fov / cell_deg))
    while npix % 2:
        npix += 1
        npix = good_size(npix)

    nx = npix
    ny = npix

    print("Image size set to (%i, %i, %i)" % (nchan, nx, ny))

    # model
    model = np.zeros((nchan, nx, ny), dtype=np.float64)
    nsource = 10
    Ix = np.random.randint(0, npix, nsource)
    Iy = np.random.randint(0, npix, nsource)
    alpha = -0.7 + 0.1 * np.random.randn(nsource)
    I0 = 1.0 + np.abs(np.random.randn(nsource))
    for i in range(nsource):
        model[:, Ix[i], Iy[i]] = I0[i] * (freq/freq0) ** alpha[i]

    # model vis
    epsilon = 1e-7
    from ducc0.wgridder import dirty2ms
    model_vis = np.zeros((nrow, nchan, ncorr), dtype=np.complex128)
    for c in range(nchan):
        model_vis[:, c:c+1, 0] = dirty2ms(uvw, freq[c:c+1], model[c],
                                    pixsize_x=cell_rad, pixsize_y=cell_rad,
                                    epsilon=epsilon, do_wstacking=True, nthreads=8)
        model_vis[:, c, -1] = model_vis[:, c, 0]


    if do_gains:
        t = (utime-utime.min())/(utime.max() - utime.min())
        nu = 2.5*(freq/freq0 - 1.0)

        from africanus.gps.utils import abs_diff
        tt = abs_diff(t, t)
        lt = 0.25
        Kt = 0.1 * np.exp(-tt**2/(2*lt**2))
        Lt = np.linalg.cholesky(Kt + 1e-10*np.eye(ntime))
        vv = abs_diff(nu, nu)
        lv = 0.1
        Kv = 0.1 * np.exp(-vv**2/(2*lv**2))
        Lv = np.linalg.cholesky(Kv + 1e-10*np.eye(nchan))
        L = (Lt, Lv)

        from pfb.utils.misc import kron_matvec
        jones = np.zeros((ntime, nchan, nant, 1, 2), dtype=np.complex128)
        for p in range(nant):
            for c in [0, -1]:  # for now only diagonal
                xi_amp = np.random.randn(ntime, nchan)
                amp = np.exp(-nu[None, :]**2 + kron_matvec(L, xi_amp))
                xi_phase = np.random.randn(ntime, nchan)
                phase = kron_matvec(L, xi_phase)
                jones[:, :, p, 0, c] = amp * np.exp(1.0j * phase)

        # corrupted vis
        model_vis = model_vis.reshape(nrow, nchan, 1, 2, 2)
        from pfb.utils.misc import chunkify_rows
        time = xds.TIME.values
        row_chunks, tbin_idx, tbin_counts = chunkify_rows(time, ntime)
        ant1 = xds.ANTENNA1.values
        ant2 = xds.ANTENNA2.values

        from africanus.calibration.utils import corrupt_vis
        gains = np.swapaxes(jones, 1, 2).copy()
        vis = corrupt_vis(tbin_idx, tbin_counts, ant1, ant2,
                          gains, model_vis).reshape(nrow, nchan, ncorr)


        xds['DATA'] = (('row','chan','corr'),
                       da.from_array(vis, chunks=(-1,-1,-1)))
        dask.compute(xds_to_table(xds, ms_name, columns='DATA'))

        # cast gain to QuartiCal format
        g = da.from_array(jones)
        gflags = da.zeros((ntime, nchan, nant, 1))
        data_vars = {
            'gains':(('gain_time', 'gain_freq', 'antenna', 'direction', 'correlation'), g),
            'gain_flags':(('gain_time', 'gain_freq', 'antenna', 'direction'), gflags)
        }
        gain_spec_tup = namedtuple('gains_spec_tup', 'tchunk fchunk achunk dchunk cchunk')
        attrs = {
            'NAME': 'NET',
            'TYPE': 'complex',
            'FIELD_NAME': '00',
            'SCAN_NUMBER': int(1),
            'FIELD_ID': int(0),
            'DATA_DESC_ID': int(0),
            'GAIN_SPEC': gain_spec_tup(tchunk=(int(ntime),),
                                       fchunk=(int(nchan),),
                                       achunk=(int(nant),),
                                       dchunk=(int(1),),
                                       cchunk=(int(2),)),
            'GAIN_AXES': ('gain_time', 'gain_freq', 'antenna', 'direction', 'correlation')
        }
        coords = {
            'gain_freq': (('gain_freq',), freq),
            'gain_time': (('gain_time',), utime)

        }
        net_xds_list = Dataset(data_vars, coords=coords, attrs=attrs)
        gain_path = str(test_dir / Path("gains.qc"))
        out_path = f'{gain_path}::NET'
        dask.compute(xds_to_zarr(net_xds_list, out_path))

    else:
        xds['DATA'] = (('row','chan','corr'),
                       da.from_array(model_vis, chunks=(-1,-1,-1)))
        dask.compute(xds_to_table(xds, ms_name, columns='DATA'))
        gain_path = None

    postfix = "main"
    # set defaults from schema
    from pfb.parser.schemas import schema
    init_args = {}
    for key in schema.init["inputs"].keys():
        init_args[key] = schema.init["inputs"][key]["default"]
    # overwrite defaults
    outname = str(test_dir / 'test')
    init_args["ms"] = str(test_dir / 'test_ascii_1h60.0s.MS')
    init_args["output_filename"] = outname
    init_args["data_column"] = "DATA"
    init_args["flag_column"] = 'FLAG'
    init_args["gain_table"] = gain_path
    init_args["max_field_of_view"] = fov
    init_args["overwrite"] = True
    init_args["channels_per_image"] = 1
    from pfb.workers.init import _init
    _init(**init_args)

    # grid data to produce dirty image
    grid_args = {}
    for key in schema.grid["inputs"].keys():
        grid_args[key] = schema.grid["inputs"][key]["default"]
    # overwrite defaults
    grid_args["output_filename"] = outname
    grid_args["postfix"] = postfix
    grid_args["nband"] = nchan
    grid_args["field_of_view"] = fov
    grid_args["fits_mfs"] = True
    grid_args["psf"] = True
    grid_args["residual"] = False
    grid_args["nthreads"] = 8  # has to be set when calling _grid
    grid_args["nvthreads"] = 8
    grid_args["overwrite"] = True
    grid_args["robustness"] = 0.0
    grid_args["wstack"] = True
    from pfb.workers.grid import _grid
    _grid(**grid_args)

    # run clean
    clean_args = {}
    for key in schema.clean["inputs"].keys():
        clean_args[key] = schema.clean["inputs"][key]["default"]
    clean_args["output_filename"] = outname
    clean_args["postfix"] = postfix
    clean_args["nband"] = nchan
    clean_args["dirosion"] = 0
    clean_args["do_residual"] = False
    clean_args["nmiter"] = 100
    threshold = 1e-5
    clean_args["threshold"] = threshold
    clean_args["gamma"] = 0.1
    clean_args["peak_factor"] = 0.75
    clean_args["sub_peak_factor"] = 0.75
    clean_args["nthreads"] = 8
    clean_args["nvthreads"] = 8
    clean_args["scheduler"] = 'sync'
    clean_args["wstack"] = True
    clean_args["epsilon"] = epsilon
    clean_args["mop_flux"] = True
    clean_args["fits_mfs"] = False
    from pfb.workers.clean import _clean
    _clean(**clean_args)

    # get inferred model
    basename = f'{outname}_I'
    dds_name = f'{basename}_{postfix}.dds'
    dds = xds_from_zarr(dds_name, chunks={'x':-1, 'y': -1})
    model_inferred = np.zeros((nchan, nx, ny))
    for ds in dds:
        b = ds.bandid
        model_inferred[b] = ds.MODEL.values

    # we actually reconstruct I/n(l,m) so we need to correct for that
    l, m = np.meshgrid(dds[0].x.values, dds[0].y.values,
                       indexing='ij')
    eps = l**2+m**2
    n = -eps/(np.sqrt(1.-eps)+1.) + 1  # more stable form
    for i in range(nsource):
        assert_allclose(1.0 + model_inferred[:, Ix[i], Iy[i]] * n[Ix[i], Iy[i]] -
                        model[:, Ix[i], Iy[i]], 1.0,
                        atol=5*threshold)



# flake8: noqa
from contextlib import ExitStack
from pfb.workers.main import cli
from functools import partial
import click
from omegaconf import OmegaConf
import pyscilog
pyscilog.init('pfb')
log = pyscilog.get_logger('FORWARD')

from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

# create default parameters from schema
defaults = {}
for key in schema.forward["inputs"].keys():
    defaults[key] = schema.forward["inputs"][key]["default"]

@cli.command(context_settings={'show_default': True})
@clickify_parameters(schema.forward)
def forward(**kw):
    '''
    Forward step aka flux mop.

    Solves

    x = (A.H R.H W R A + sigmainv**2 I)^{-1} ID

    with a suitable approximation to R.H W R (eg. convolution with the PSF,
    BDA'd weights or none). Here A is the combination of mask and an
    average beam pattern.

    If a host address is provided the computation can be distributed
    over imaging band and row. When using a distributed scheduler both
    mem-limit and nthreads is per node and have to be specified.

    When using a local cluster, mem-limit and nthreads refer to the global
    memory and threads available, respectively. By default the gridder will
    use all available resources.

    On a local cluster, the default is to use:

        nworkers = nband
        nthreads-per-worker = 1

    They have to be specified in ~.config/dask/jobqueue.yaml in the
    distributed case.

    if LocalCluster:
        nvthreads = nthreads//(nworkers*nthreads_per_worker)
    else:
        nvthreads = nthreads//nthreads-per-worker

    where nvthreads refers to the number of threads used to scale vertically
    (eg. the number threads given to each gridder instance).

    '''
    defaults.update(kw)
    opts = OmegaConf.create(defaults)
    import time
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    pyscilog.log_to_file(f'forward_{timestamp}.log')

    if opts.nworkers is None:
        if opts.scheduler=='distributed':
            opts.nworkers = opts.nband
        else:
            opts.nworkers = 1

    OmegaConf.set_struct(opts, True)

    with ExitStack() as stack:
        from pfb import set_client
        opts = set_client(opts, stack, log, scheduler=opts.scheduler)

        # TODO - prettier config printing
        print('Input Options:', file=log)
        for key in opts.keys():
            print('     %25s = %s' % (key, opts[key]), file=log)

        return _forward(**opts)

def _forward(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    import numpy as np
    import numexpr as ne
    import xarray as xr
    import dask
    import dask.array as da
    from daskms.experimental.zarr import xds_from_zarr, xds_to_zarr
    from pfb.utils.fits import load_fits, dds2fits_mfs, dds2fits
    from pfb.utils.misc import init_mask, dds2cubes
    from pfb.operators.hessian import hessian_xds, hessian_psf_cube
    from pfb.opt.pcg import pcg
    from ducc0.misc import make_noncritical

    basename = f'{opts.output_filename}_{opts.product.upper()}'

    dds_name = f'{basename}_{opts.postfix}.dds'

    dds = xds_from_zarr(dds_name, chunks={'row':-1,
                                          'chan': -1})

    # stitch image space data products
    output_type = dds[0].DIRTY.dtype
    print("Combining slices into cubes", file=log)
    dirty, model, residual, psf, psfhat, beam, wsums, dual = dds2cubes(
                                                               dds,
                                                               opts.nband,
                                                               apparent=False)
    wsum = np.sum(wsums)
    psf_mfs = np.sum(psf, axis=0)
    assert (psf_mfs.max() - 1.0) < 2*opts.epsilon
    dirty_mfs = np.sum(dirty, axis=0)
    if residual is None:
        residual = dirty.copy()
        residual_mfs = dirty_mfs.copy()
    else:
        residual_mfs = np.sum(residual, axis=0)

    lastsize = dds[0].y_psf.size

    # for intermediary results (not currently written)
    freq_out = []
    for ds in dds:
        freq_out.append(ds.freq_out)
    freq_out = np.unique(np.array(freq_out))
    nband = opts.nband
    nx = dds[0].x.size
    ny = dds[0].y.size
    ra = dds[0].ra
    dec = dds[0].dec
    radec = [ra, dec]
    cell_rad = dds[0].cell_rad
    cell_deg = np.rad2deg(cell_rad)
    ref_freq = np.mean(freq_out)

    if opts.mask is not None:
        mask = load_fits(opts.mask, dtype=output_type).squeeze()
        assert mask.shape == (nx, ny)
        print('Using provided fits mask', file=log)
    else:
        mask = np.ones((nx, ny), dtype=dirty.dtype)

    # set up vis space Hessian for computing the residual
    # TODO - how to apply beam externally per ds
    hessopts = {}
    hessopts['cell'] = dds[0].cell_rad
    hessopts['wstack'] = opts.wstack
    hessopts['epsilon'] = opts.epsilon
    hessopts['double_accum'] = opts.double_accum
    hessopts['nthreads'] = opts.nvthreads
    hess = partial(hessian_xds, xds=dds, hessopts=hessopts,
                   wsum=wsum, sigmainv=0,
                   mask=np.ones((nx, ny), dtype=output_type),
                   compute=True, use_beam=False)

    if opts.use_psf:
        print("Solving for update using image space hessian approximation",
              file=log)
        xout = np.empty(dirty.shape, dtype=dirty.dtype, order='C')
        xout = make_noncritical(xout)
        xpad = np.empty(psf.shape, dtype=dirty.dtype, order='C')
        xpad = make_noncritical(xpad)
        xhat = np.empty(psfhat.shape, dtype=psfhat.dtype)
        xhat = make_noncritical(xhat)
        hess_pcg = partial(hessian_psf_cube,
                           xpad,
                           xhat,
                           xout,
                           beam*mask[None, :, :],
                           psfhat,
                           lastsize,
                           nthreads=opts.nvthreads*opts.nthreads_dask,  # not using dask parallelism
                           sigmainv=opts.sigmainv)
    else:
        print("Solving for update using vis space hessian approximation",
              file=log)
        hess_pcg = partial(hessian_xds, xds=dds, hessopts=hessopts,
                   wsum=wsum, sigmainv=opts.sigmainv,
                   mask=mask,
                   compute=True, use_beam=False)

    cgopts = {}
    cgopts['tol'] = opts.cg_tol
    cgopts['maxit'] = opts.cg_maxit
    cgopts['minit'] = opts.cg_minit
    cgopts['verbosity'] = opts.cg_verbose
    cgopts['report_freq'] = opts.cg_report_freq
    cgopts['backtrack'] = opts.backtrack

    print("Solving for update", file=log)
    x0 = np.zeros((nband, nx, ny), dtype=output_type)
    update = pcg(hess_pcg, beam * mask[None, :, :] * residual, x0,
                 tol=opts.cg_tol,
                 maxit=opts.cg_maxit,
                 minit=opts.cg_minit,
                 verbosity=opts.cg_verbose,
                 report_freq=opts.cg_report_freq,
                 backtrack=opts.backtrack)

    modelp = model.copy()
    model += opts.gamma * update


    print("Getting residual", file=log)
    convimage = hess(model)
    ne.evaluate('dirty - convimage', out=residual,
                casting='same_kind')
    ne.evaluate('sum(residual, axis=0)', out=residual_mfs,
                casting='same_kind')


    print("Updating results", file=log)
    dds_out = []
    for ds in dds:
        b = ds.bandid
        r = da.from_array(residual[b]*wsum)
        m = da.from_array(model[b])
        mp = da.from_array(modelp[b])
        u = da.from_array(update[b])
        ds_out = ds.assign(**{'RESIDUAL': (('x', 'y'), r),
                              'MODEL': (('x', 'y'), m),
                              'MODELP': (('x', 'y'), mp),  # to revert in case of failure
                              'UPDATE': (('x', 'y'), u)})
        dds_out.append(ds_out)
    writes = xds_to_zarr(dds_out, dds_name,
                         columns=('RESIDUAL', 'MODEL', 'MODELP', 'DUAL'),
                         rechunk=True)
    dask.compute(writes)

    dds = xds_from_zarr(dds_name, chunks={'x': -1, 'y': -1})

    # convert to fits files
    fitsout = []
    if opts.fits_mfs:
        fitsout.append(dds2fits_mfs(dds,
                                    'RESIDUAL',
                                    f'{basename}_{opts.postfix}',
                                    norm_wsum=True))
        fitsout.append(dds2fits_mfs(dds,
                                    'MODEL',
                                    f'{basename}_{opts.postfix}',
                                    norm_wsum=False))

    if opts.fits_cubes:
        fitsout.append(dds2fits(dds,
                                'RESIDUAL',
                                f'{basename}_{opts.postfix}',
                                norm_wsum=True))
        fitsout.append(dds2fits(dds,
                                'MODEL',
                                f'{basename}_{opts.postfix}',
                                norm_wsum=False))

    if len(fitsout):
        print("Writing fits", file=log)
        dask.compute(fitsout)

    print("All done here.", file=log)



from dataclasses import dataclass
import os.path
import glob
from typing import *
from scabha import configuratt
from scabha.cargo import Parameter
from omegaconf.omegaconf import OmegaConf


schema = None

@dataclass
class _CabInputsOutputs(object):
    inputs: Dict[str, Parameter]
    outputs: Dict[str, Parameter]

# load schema files
if schema is None:

    # all *.yaml files under pfb.parser will be loaded automatically

    files = glob.glob(os.path.join(os.path.dirname(__file__), "*.yaml"))

    structured = OmegaConf.structured(_CabInputsOutputs)

    tmp = configuratt.load_nested(files,
                                  structured=structured,
                                  config_class="PfbCleanCabs",
                                  use_cache=False)

    # this is required since upgrade of scabha to caching branch
    # tmp is a tuple containing the config object as the first element
    # and a set containing locations of .yaml configs for pfb workers
    schema = OmegaConf.create(tmp[0])

    for worker in schema.keys():
        for param in schema[worker]['inputs']:
            if schema[worker]['inputs'][param]['default'] == '<UNSET DEFAULT VALUE>':
                schema[worker]['inputs'][param]['default'] = None






import pytest
from pathlib import Path
from pfb.utils.weighting import (_compute_counts, _counts_to_weights,
                                 compute_counts, counts_to_weights)

pmp = pytest.mark.parametrize

def test_counts(ms_name):
    '''
    Compares _compute_counts to memory greedy numpy implementation
    '''

    import numpy as np
    np.random.seed(420)
    from numpy.testing import assert_allclose
    from daskms import xds_from_ms, xds_from_table
    from africanus.constants import c as lightspeed

    test_dir = Path(ms_name).resolve().parent
    xds = xds_from_ms(ms_name,
                      chunks={'row': -1, 'chan': -1})[0]
    spw = xds_from_table(f'{ms_name}::SPECTRAL_WINDOW')[0]

    utime = np.unique(xds.TIME.values)
    freq = spw.CHAN_FREQ.values.squeeze()
    freq0 = np.mean(freq)

    ntime = utime.size
    nchan = freq.size
    nant = np.maximum(xds.ANTENNA1.values.max(), xds.ANTENNA1.values.max()) + 1

    ncorr = xds.corr.size

    uvw = xds.UVW.values
    nrow = uvw.shape[0]
    u_max = abs(uvw[:, 0]).max()
    v_max = abs(uvw[:, 1]).max()
    uv_max = np.maximum(u_max, v_max)

    # image size
    cell_N = 1.0 / (2 * uv_max * freq.max() / lightspeed)

    srf = 2.0
    cell_rad = cell_N / srf
    cell_deg = cell_rad * 180 / np.pi
    cell_size = cell_deg * 3600
    print("Cell size set to %5.5e arcseconds" % cell_size)

    from ducc0.fft import good_size
    # the test will fail in intrinsic if sources fall near beam sidelobes
    fov = 1.0
    npix = good_size(int(fov / cell_deg))
    while npix % 2:
        npix += 1
        npix = good_size(npix)

    nx = npix
    ny = npix

    print("Image size set to (%i, %i, %i)" % (nchan, nx, ny))


    mask = np.ones((nrow, nchan), dtype=bool)
    counts = _compute_counts(uvw, freq, mask, nx, ny, cell_rad, cell_rad,
                             dtype=np.float64, k=0).squeeze()
    ku = np.sort(np.fft.fftfreq(nx, cell_rad))
    # shift by half a pixel to get bin edges
    kucell = ku[1] - ku[0]
    ku -= kucell/2
    # add upper edge
    ku = np.append(ku, ku.max() + kucell)
    kv = np.sort(np.fft.fftfreq(ny, cell_rad))
    kvcell = kv[1] - kv[0]
    kv -= kvcell/2
    kv = np.append(kv, kv.max() + kvcell)
    weights = np.ones((nrow*nchan), dtype=np.float64)
    u = (uvw[:, 0:1] * freq[None, :]/lightspeed).ravel()
    v = (uvw[:, 1:2] * freq[None, :]/lightspeed).ravel()
    counts2, _, _ = np.histogram2d(u, v, bins=[ku, kv], weights=weights)

    assert_allclose(counts, counts2)


def test_counts_dask(ms_name):
    '''
    Compares dask compute_counts to numba implementation
    '''

    import numpy as np
    np.random.seed(420)
    from numpy.testing import assert_allclose
    from daskms import xds_from_ms, xds_from_table
    from africanus.constants import c as lightspeed
    import dask.array as da
    import dask
    dask.config.set(scheduler='sync')

    test_dir = Path(ms_name).resolve().parent
    xds = xds_from_ms(ms_name,
                      chunks={'row': -1, 'chan': -1})[0]
    spw = xds_from_table(f'{ms_name}::SPECTRAL_WINDOW')[0]

    utime = np.unique(xds.TIME.values)
    freq = spw.CHAN_FREQ.values.squeeze()
    freq0 = np.mean(freq)

    ntime = utime.size
    nchan = freq.size
    nant = np.maximum(xds.ANTENNA1.values.max(), xds.ANTENNA2.values.max()) + 1

    ncorr = xds.corr.size

    uvw = xds.UVW.values
    nrow = uvw.shape[0]
    u_max = abs(uvw[:, 0]).max()
    v_max = abs(uvw[:, 1]).max()
    uv_max = np.maximum(u_max, v_max)

    # image size
    cell_N = 1.0 / (2 * uv_max * freq.max() / lightspeed)

    srf = 2.0
    cell_rad = cell_N / srf
    cell_deg = cell_rad * 180 / np.pi
    cell_size = cell_deg * 3600
    print("Cell size set to %5.5e arcseconds" % cell_size)

    from ducc0.fft import good_size
    # the test will fail in intrinsic if sources fall near beam sidelobes
    fov = 1.0
    npix = good_size(int(fov / cell_deg))
    while npix % 2:
        npix += 1
        npix = good_size(npix)

    nx = npix
    ny = npix

    print("Image size set to (%i, %i, %i)" % (nchan, nx, ny))
    mask = np.ones((nrow, nchan), dtype=bool)
    counts = _compute_counts(uvw, freq, mask, nx, ny, cell_rad, cell_rad,
                             np.float64, k=0).squeeze()

    rc = 5000
    uvw = da.from_array(uvw, chunks=(rc, 3))
    freq = da.from_array(freq, chunks=-1)
    mask = da.from_array(mask, chunks=(rc, -1))
    counts_dask = compute_counts(uvw, freq, mask, nx, ny, cell_rad, cell_rad,
                                 np.float64, k=0)

    assert_allclose(counts, counts_dask.compute())


# deprecated
# def test_uniform(tmp_path_factory):
#     '''
#     Tests that the grid is uniformly weighted after doing weighting
#     '''
#     test_dir = tmp_path_factory.mktemp("test_weighting")
#     # test_dir = Path('/home/landman/data/')
#     packratt.get('/test/ms/2021-06-24/elwood/test_ascii_1h60.0s.MS.tar', str(test_dir))

#     import numpy as np
#     np.random.seed(420)
#     from numpy.testing import assert_allclose
#     from pyrap.tables import table

#     ms = table(str(test_dir / 'test_ascii_1h60.0s.MS'), readonly=False)
#     freq = table(str(test_dir / 'test_ascii_1h60.0s.MS::SPECTRAL_WINDOW')).getcol('CHAN_FREQ').squeeze()
#     nchan = freq.size
#     uvw = ms.getcol('UVW')
#     nrow = uvw.shape[0]
#     u_max = abs(uvw[:, 0]).max()
#     v_max = abs(uvw[:, 1]).max()
#     uv_max = np.maximum(u_max, v_max)

#     # image size
#     from africanus.constants import c as lightspeed
#     cell_N = 1.0 / (2 * uv_max * freq.max() / lightspeed)

#     srf = 2.0
#     cell_rad = cell_N / srf
#     cell_deg = cell_rad * 180 / np.pi
#     cell_size = cell_deg * 3600

#     from ducc0.fft import good_size
#     fov = 1.0
#     npix = good_size(int(fov / cell_deg))
#     while npix % 2:
#         npix += 1
#         npix = good_size(npix)

#     nx = npix
#     ny = npix

#     mask = np.ones((nrow, nchan), dtype=bool)
#     counts = _compute_counts(uvw, freq, mask, nx, ny, cell_rad, cell_rad,
#                              dtype=np.float64, k=0).squeeze()

#     weights = _counts_to_weights(counts, uvw, freq, nx, ny,
#                                  cell_rad, cell_rad, -3)


#     counts2 = _compute_counts(uvw, freq, mask, nx, ny, cell_rad, cell_rad,
#                               np.float64, wgt=weights, k=0).squeeze()

#     assert_allclose(counts2[counts2>0], 1)


# def test_uniform_dask(tmp_path_factory):
#     test_dir = tmp_path_factory.mktemp("test_weighting")
#     # test_dir = Path('/home/landman/data/')
#     packratt.get('/test/ms/2021-06-24/elwood/test_ascii_1h60.0s.MS.tar', str(test_dir))

#     import numpy as np
#     np.random.seed(420)
#     from numpy.testing import assert_allclose
#     from pyrap.tables import table
#     import dask.array as da
#     import dask

#     ms = table(str(test_dir / 'test_ascii_1h60.0s.MS'), readonly=False)
#     freq = table(str(test_dir / 'test_ascii_1h60.0s.MS::SPECTRAL_WINDOW')).getcol('CHAN_FREQ').squeeze()
#     nchan = freq.size
#     uvw = ms.getcol('UVW')
#     nrow = uvw.shape[0]
#     u_max = abs(uvw[:, 0]).max()
#     v_max = abs(uvw[:, 1]).max()
#     uv_max = np.maximum(u_max, v_max)

#     # image size
#     from africanus.constants import c as lightspeed
#     cell_N = 1.0 / (2 * uv_max * freq.max() / lightspeed)

#     srf = 2.0
#     cell_rad = cell_N / srf
#     cell_deg = cell_rad * 180 / np.pi
#     cell_size = cell_deg * 3600

#     from ducc0.fft import good_size
#     fov = 1.0
#     npix = good_size(int(fov / cell_deg))
#     while npix % 2:
#         npix += 1
#         npix = good_size(npix)

#     nx = npix
#     ny = npix

#     rc = 5000
#     uvw = da.from_array(uvw, chunks=(rc, 3))
#     freq = da.from_array(freq, chunks=-1)
#     mask = da.ones((nrow, nchan), chunks=(rc, -1), dtype=bool)
#     counts = compute_counts(uvw, freq, mask, nx, ny, cell_rad, cell_rad,
#                             np.float64, k=0)

#     weights = counts_to_weights(counts, uvw, freq, nx, ny,
#                                 cell_rad, cell_rad, -3)

#     counts2 = compute_counts(uvw, freq, mask, nx, ny, cell_rad, cell_rad,
#                              np.float64, wgt=weights, k=0)

#     counts2 = counts2.compute()
#     assert_allclose(counts2[counts2>0], 1)



# flake8: noqa
from contextlib import ExitStack
from pfb.workers.main import cli
import click
from omegaconf import OmegaConf
import pyscilog
pyscilog.init('pfb')
log = pyscilog.get_logger('GRID')

from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

# create default parameters from schema
defaults = {}
for key in schema.grid["inputs"].keys():
    defaults[key] = schema.grid["inputs"][key]["default"]

@cli.command(context_settings={'show_default': True})
@clickify_parameters(schema.grid)
def grid(**kw):
    '''
    Compute imaging weights and create a dirty image, psf etc.
    By default only the MFS images are converted to fits files.
    Set the --fits-cubes flag to also produce fits cubes.

    '''
    defaults.update(kw)
    opts = OmegaConf.create(defaults)
    import time
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    pyscilog.log_to_file(f'grid_{timestamp}.log')

    OmegaConf.set_struct(opts, True)

    with ExitStack() as stack:
        from pfb import set_client
        opts = set_client(opts, stack, log, scheduler=opts.scheduler)

        # TODO - prettier config printing
        print('Input Options:', file=log)
        for key in opts.keys():
            print('     %25s = %s' % (key, opts[key]), file=log)

        return _grid(**opts)

def _grid(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    import os
    import numpy as np
    import dask
    dask.config.set(**{'array.slicing.split_large_chunks': False})
    from daskms.experimental.zarr import xds_to_zarr, xds_from_zarr
    import dask.array as da
    from africanus.constants import c as lightspeed
    from ducc0.fft import good_size
    from pfb.utils.fits import dds2fits, dds2fits_mfs
    from pfb.utils.misc import compute_context
    from pfb.operators.gridder import image_data_products
    from pfb.operators.fft import fft2d
    from pfb.utils.weighting import (compute_counts,
                                     filter_extreme_counts)
    from pfb.utils.beam import eval_beam
    import xarray as xr
    from uuid import uuid4
    from dask.graph_manipulation import clone
    from pfb.utils.astrometry import get_coordinates
    from africanus.coordinates import radec_to_lm
    from pfb.utils.misc import concat_chan, concat_row
    import sympy as sm
    from sympy.utilities.lambdify import lambdify
    from sympy.parsing.sympy_parser import parse_expr
    from quartical.utils.dask import Blocker


    basename = f'{opts.output_filename}_{opts.product.upper()}'

    # xds contains vis products, no imaging weights applied
    xds_name = f'{basename}.xds'
    xds = xds_from_zarr(xds_name, chunks={'row': -1})
    # dds contains image space products including imaging weights and uvw
    dds_name = f'{basename}_{opts.postfix}.dds'

    if os.path.isdir(dds_name):
        dds_exists = True
        if opts.overwrite:
            print(f'Removing {dds_name}', file=log)
            import shutil
            shutil.rmtree(dds_name)
            dds_exists = False
    else:
        dds_exists = False

    times_in = []
    freqs_in = []
    for ds in xds:
        times_in.append(ds.time_out)
        freqs_in.append(ds.freq_out)

    times_in = np.unique(times_in)
    freqs_in = np.unique(freqs_in)

    ntime_in = times_in.size
    nband_in = freqs_in.size

    if opts.concat_row and len(xds) > nband_in:
        print('Concatenating datasets along row dimension', file=log)
        xds = concat_row(xds)
        ntime = 1
        times_out = np.array((xds[0].time_out,))
    else:
        ntime = ntime_in
        times_out = times_in

    if opts.nband != nband_in and len(xds) > ntime:
        print('Concatenating datasets along chan dimension. '
              f'Mapping {nband_in} datasets to {opts.nband} bands', file=log)
        xds = concat_chan(xds, nband_out=opts.nband)
        nband = opts.nband
        freqs_out = []
        for ds in xds:
            freqs_out.append(ds.freq_out)
        freqs_out = np.unique(freqs_out)
    else:
        nband = nband_in
        freqs_out = freqs_in

    # do this after concatenation (to check)
    for i, ds in enumerate(xds):
        xds[i] = ds.chunk({'chan': -1})

    real_type = xds[0].WEIGHT.dtype
    if real_type == np.float32:
        precision = 'single'
    else:
        precision = 'double'

    # max uv coords over all datasets
    uv_maxs = []
    max_freqs = []
    for ds in xds:
        uvw = ds.UVW.data
        u_max = abs(uvw[:, 0]).max()
        v_max = abs(uvw[:, 1]).max()
        uv_maxs.append(da.maximum(u_max, v_max))
        max_freqs.append(ds.FREQ.data.max())

    # early compute necessary to set image size
    uv_maxs, max_freqs = dask.compute(uv_maxs, max_freqs)
    uv_max = max(uv_maxs)
    max_freq = max(max_freqs)

    # max cell size
    cell_N = 1.0 / (2 * uv_max * max_freq / lightspeed)

    if opts.cell_size is not None:
        cell_size = opts.cell_size
        cell_rad = cell_size * np.pi / 60 / 60 / 180
        if cell_N / cell_rad < 1:
            raise ValueError("Requested cell size too large. "
                             "Super resolution factor = ", cell_N / cell_rad)
        print(f"Super resolution factor = {cell_N/cell_rad}", file=log)
    else:
        cell_rad = cell_N / opts.super_resolution_factor
        cell_size = cell_rad * 60 * 60 * 180 / np.pi
        print(f"Cell size set to {cell_size} arcseconds", file=log)

    if opts.nx is None:
        fov = opts.field_of_view * 3600
        npix = int(fov / cell_size)
        npix = good_size(npix)
        while npix % 2:
            npix += 1
            npix = good_size(npix)
        nx = npix
        ny = npix
    else:
        nx = opts.nx
        ny = opts.ny if opts.ny is not None else nx

    if opts.dirty:
        print(f"Image size = (ntime={ntime}, nband={nband}, nx={nx}, ny={ny})", file=log)

    nx_psf = good_size(int(opts.psf_oversize * nx))
    while nx_psf % 2:
        nx_psf += 1
        nx_psf = good_size(nx_psf)

    ny_psf = good_size(int(opts.psf_oversize * ny))
    while ny_psf % 2:
        ny_psf += 1
        ny_psf = good_size(ny_psf)
    nyo2 = ny_psf//2 + 1

    if opts.psf:
        print(f"PSF size = (ntime={ntime}, nband={nband}, nx={nx_psf}, ny={ny_psf})", file=log)

    # if dds exists, check that existing dds is compatible with input
    if dds_exists:
        dds = xds_from_zarr(dds_name)
        # these need to be aligned at this stage
        for ds, out_ds in zip(xds, dds):
            assert ds.freq_out == out_ds.freq_out
            assert ds.time_out == out_ds.time_out
            assert ds.ra == out_ds.ra
            assert ds.dec == out_ds.dec
            assert out_ds.x.size == nx
            assert out_ds.y.size == ny
            assert out_ds.cell_rad == cell_rad
        print(f'As far as we can tell {dds_name} can be reused/updated.',
              file=log)
    else:
        print(f'Image space data products will be stored in {dds_name}.', file=log)

    # check if model exists
    if opts.transfer_model_from is not None:
        try:
            mds = xds_from_zarr(opts.transfer_model_from,
                                chunks={'params':-1, 'comps':-1})[0]
        except Exception as e:
            raise ValueError(f"No dataset found at {opts.transfer_model_from}")

        # should we construct the model func outside
        # of eval_coeffs_to_slice?
        # params = sm.symbols(('t','f'))
        # params += sm.symbols(tuple(mds.params.values))
        # symexpr = parse_expr(mds.parametrisation)
        # model_func = lambdify(params, symexpr)
        # texpr = parse_expr(mds.texpr)
        # tfunc = lambdify(params[0], texpr)
        # fexpr = parse_expr(mds.fexpr)
        # ffunc = lambdify(params[1], fexpr)


        # we only want to load these once
        model_coeffs = mds.coefficients.values
        locx = mds.location_x.values
        locy = mds.location_y.values
        params = mds.params.values
        coeffs = mds.coefficients.values

        print(f"Loading model from {opts.transfer_model_from}. ",
              file=log)

    dds_out = []
    for i, ds in enumerate(xds):
        if dds_exists:
            out_ds = dds[i].chunk({'row':-1,
                                   'chan':-1,
                                   'x':-1,
                                   'y':-1})
        else:
            out_ds = xr.Dataset()
        uvw = ds.UVW.data
        freq = ds.FREQ.data
        vis = ds.VIS.data
        wgt = ds.WEIGHT.data
        # This is a vis space mask (see wgridder convention)
        mask = ds.MASK.data
        bandid = np.where(freqs_out == ds.freq_out)[0][0]
        timeid = np.where(times_out == ds.time_out)[0][0]

        # compute lm coordinates of target
        if opts.target is not None:
            tmp = opts.target.split(',')
            if len(tmp) == 1 and tmp[0] == opts.target:
                obs_time = ds.time_out
                tra, tdec = get_coordinates(obs_time, target=opts.target)
            else:  # we assume a HH:MM:SS,DD:MM:SS format has been passed in
                from astropy import units as u
                from astropy.coordinates import SkyCoord
                c = SkyCoord(tmp[0], tmp[1], frame='fk5', unit=(u.hourangle, u.deg))
                tra = np.deg2rad(c.ra.value)
                tdec = np.deg2rad(c.dec.value)

            tcoords=np.zeros((1,2))
            tcoords[0,0] = tra
            tcoords[0,1] = tdec
            coords0 = np.array((ds.ra, ds.dec))
            lm0 = radec_to_lm(tcoords, coords0).squeeze()
            # LB - why the negative?
            x0 = -lm0[0]
            y0 = -lm0[1]
        else:
            x0 = 0.0
            y0 = 0.0
            tra = ds.ra
            tdec = ds.dec

        out_ds = out_ds.assign_attrs(**{
            'ra': tra,
            'dec': tdec,
            'x0': x0,
            'y0': y0,
            'cell_rad': cell_rad,
            'bandid': bandid,
            'timeid': timeid,
            'freq_out': ds.freq_out,
            'time_out': ds.time_out,
            'robustness': opts.robustness
        })
        # TODO - assign ug,vg-coordinates
        x = (-nx/2 + np.arange(nx)) * cell_rad + x0
        y = (-ny/2 + np.arange(ny)) * cell_rad + y0
        out_ds = out_ds.assign_coords(**{
           'x': x,
           'y': y
        })

        # evaluate beam at x and y coords
        cell_deg = np.rad2deg(cell_rad)
        l = (-(nx//2) + da.arange(nx)) * cell_deg + np.deg2rad(x0)
        m = (-(ny//2) + da.arange(ny)) * cell_deg + np.deg2rad(y0)
        ll, mm = da.meshgrid(l, m, indexing='ij')
        bvals = eval_beam(ds.BEAM.data, ll, mm)
        out_ds = out_ds.assign(**{'BEAM': (('x', 'y'), bvals)})

        # get the model
        if opts.transfer_model_from is not None:
            from pfb.utils.misc import eval_coeffs_to_slice
            model = eval_coeffs_to_slice(
                ds.time_out,
                ds.freq_out,
                model_coeffs,
                locx, locy,
                mds.parametrisation,
                params,
                mds.texpr,
                mds.fexpr,
                mds.npix_x, mds.npix_y,
                mds.cell_rad_x, mds.cell_rad_y,
                mds.center_x, mds.center_y,
                nx, ny,
                cell_rad, cell_rad,
                x0, y0
            )
            model = da.from_array(model, chunks=(-1,-1))
            out_ds = out_ds.assign(**{'MODEL': (('x', 'y'), model)})

        elif 'MODEL' in out_ds:
            model = out_ds.MODEL.data
        else:
            model = None


        if opts.robustness is not None:
            # we'll skip this if counts already exists
            # what to do if flags have changed?
            if 'COUNTS' not in out_ds:
                counts = compute_counts(
                        clone(uvw),
                        freq,
                        mask,
                        nx,
                        ny,
                        cell_rad,
                        cell_rad,
                        real_type,
                        ngrid=opts.nvthreads)
                # get rid of artificially high weights corresponding to
                # nearly empty cells
                if opts.filter_extreme_counts:
                    counts = filter_extreme_counts(counts, nbox=opts.filter_nbox,
                                                   nlevel=opts.filter_level)

                # do we want the coordinates to be ug, vg rather?
                out_ds = out_ds.assign(**{'COUNTS': (('x', 'y'), counts)})
        else:
            counts = None

        # we might want to chunk over row chan in the future but
        # for now these will always have chunks=(-1,-1)
        blocker = Blocker(image_data_products, ('row', 'chan'))
        blocker.add_input('uvw', uvw, ('row','three'))
        blocker.add_input('freq', freq, ('chan',))
        blocker.add_input('vis', vis, ('row','chan'))
        blocker.add_input('wgt', wgt, ('row','chan'))
        blocker.add_input('mask', mask, ('row','chan'))
        if counts is not None:
            blocker.add_input('counts', counts, ('x','y'))
        else:
            blocker.add_input('counts', None)
        blocker.add_input('nx', nx)
        blocker.add_input('ny', ny)
        blocker.add_input('nx_psf', nx_psf)
        blocker.add_input('ny_psf', ny_psf)
        blocker.add_input('cellx', cell_rad)
        blocker.add_input('celly', cell_rad)
        if model is not None:
            blocker.add_input('model', model, ('x', 'y'))
        else:
            blocker.add_input('model', None)
        blocker.add_input('robustness', opts.robustness)
        blocker.add_input('x0', x0)
        blocker.add_input('y0', y0)
        blocker.add_input('nthreads', opts.nvthreads)
        blocker.add_input('epsilon', opts.epsilon)
        blocker.add_input('do_wgridding', opts.wstack)
        blocker.add_input('double_accum', opts.double_accum)
        blocker.add_input('l2reweight_dof', opts.l2reweight_dof)
        blocker.add_input('do_psf', opts.psf)
        blocker.add_input('do_weight', opts.weight)
        blocker.add_input('do_residual', opts.residual)

        blocker.add_output(
            'DIRTY',
            ('x', 'y'),
            ((nx,), (ny,)),
            wgt.dtype)

        blocker.add_output(
            'WSUM',
            ('scalar',),
            ((1,),),
            wgt.dtype)

        if opts.residual:
            blocker.add_output(
                'RESIDUAL',
                ('x', 'y'),
                ((nx,), (ny,)),
                wgt.dtype)

        if opts.psf:
            blocker.add_output(
                'PSF',
                ('x_psf', 'y_psf'),
                ((nx_psf,), (ny_psf,)),
                wgt.dtype)
            blocker.add_output(
                'PSFHAT',
                ('x_psf', 'yo2'),
                ((nx_psf,), (nyo2,)),
                wgt.dtype)

        if opts.weight:
            blocker.add_output(
                'WEIGHT',
                ('row', 'chan'),
                wgt.chunks,
                wgt.dtype)

        output_dict = blocker.get_dask_outputs()
        out_ds = out_ds.assign(**{
            'DIRTY': (('x', 'y'), output_dict['DIRTY'])
            })

        if opts.psf:
            out_ds = out_ds.assign(**{
                'PSF': (('x_psf', 'y_psf'), output_dict['PSF']),
                'PSFHAT': (('x_psf', 'yo2'), output_dict['PSFHAT'])
                })

        # TODO - don't put vis space products in dds
        # but how apply imaging weights in that case?
        if opts.weight:
            out_ds = out_ds.assign(**{
                'WEIGHT': (('row', 'chan'), output_dict['WEIGHT'])
                })

        if opts.residual:
            out_ds = out_ds.assign(**{
                'RESIDUAL': (('x', 'y'), output_dict['RESIDUAL'])
                })


        out_ds = out_ds.assign(**{
            'FREQ': (('chan',), freq),
            'UVW': (('row', 'three'), uvw),
            'MASK': (('row', 'chan'), mask),
            'WSUM': (('scalar',), output_dict['WSUM'])
            })


        out_ds = out_ds.chunk({'row':100000,
                               'chan':128,
                               'x':4096,
                               'y':4096})
        # necessary to make psf optional
        if 'x_psf' in out_ds.dims:
            out_ds = out_ds.chunk({'x_psf': 4096,
                                   'y_psf':4096,
                                   'yo2': 2048})

        dds_out.append(out_ds.unify_chunks())

    writes = xds_to_zarr(dds_out, dds_name, columns='ALL')

    # dask.visualize(writes, color="order", cmap="autumn",
    #                node_attr={"penwidth": "4"},
    #                filename=f'{basename}_grid_ordered_graph.pdf',
    #                optimize_graph=False)
    # dask.visualize(writes, filename=f'{basename}_grid_graph.pdf',
    #                optimize_graph=False)

    print("Computing image space data products", file=log)
    with compute_context(opts.scheduler, basename+'_grid'):
        dask.compute(writes, optimize_graph=False)

    dds = xds_from_zarr(dds_name, chunks={'x': -1, 'y': -1})

    # convert to fits files
    fitsout = []
    if opts.fits_mfs:
        if opts.dirty:
            fitsout.append(dds2fits_mfs(dds, 'DIRTY', f'{basename}_{opts.postfix}', norm_wsum=True))
        if opts.psf:
            fitsout.append(dds2fits_mfs(dds, 'PSF', f'{basename}_{opts.postfix}', norm_wsum=True))
        if opts.residual and model is not None:
            fitsout.append(dds2fits_mfs(dds, 'RESIDUAL', f'{basename}_{opts.postfix}', norm_wsum=True))
            fitsout.append(dds2fits_mfs(dds, 'MODEL', f'{basename}_{opts.postfix}', norm_wsum=False))

    if opts.fits_cubes:
        if opts.dirty:
            fitsout.append(dds2fits(dds, 'DIRTY', f'{basename}_{opts.postfix}', norm_wsum=True))
        if opts.psf:
            fitsout.append(dds2fits(dds, 'PSF', f'{basename}_{opts.postfix}', norm_wsum=True))
        if opts.residual and model is not None:
            fitsout.append(dds2fits(dds, 'RESIDUAL', f'{basename}_{opts.postfix}', norm_wsum=True))
            fitsout.append(dds2fits(dds, 'MODEL', f'{basename}_{opts.postfix}', norm_wsum=False))

    if len(fitsout):
        print("Writing fits", file=log)
        dask.compute(fitsout)

    if opts.scheduler=='distributed':
        from distributed import get_client
        client = get_client()
        client.close()

    print("All done here.", file=log)



import numpy as np
from pfb.operators.psf import PSF
from pfb.opt.hogbom import hogbom
import pyscilog
log = pyscilog.get_logger('CLEAN')


def resid_func(x, dirty, psfo):
    residual = dirty - psfo.convolve(x)
    return residual


def clean(psf, model, residual, mask=None, beam=None,
          nthreads=0, maxit=10, gamma=1.0, peak_factor=0.01, threshold=0.0,
          hbgamma=0.1, hbpf=0.1, hbmaxit=5000, hbverbose=1):  # Hogbom options

    if len(residual.shape) > 3:
        raise ValueError("Residual must have shape (nband, nx, ny)")

    nband, nx, ny = residual.shape

    if beam is not None:
        raise NotImplementedError("Beam not supported in clean minor cycle")

    if mask is None:
        def mask(x): return x
    else:
        try:
            if mask.ndim == 2:
                assert mask.shape == (nx, ny)
                def mask(x): return mask[None] * x
            elif mask.ndim == 3:
                assert mask.shape == (1, nx, ny)
                def mask(x): return mask * x
            else:
                raise ValueError
        except BaseException:
            raise ValueError("Mask has incorrect shape")

    # PSF operator
    psfo = PSF(psf, nthreads, imsize=residual.shape)

    # init dirty
    if model.any():
        dirty = residual + psfo.convolve(model)
    else:
        dirty = residual

    # residual
    residual_mfs = np.sum(residual, axis=0)
    rms = np.std(residual_mfs)
    rmax = np.abs(residual_mfs).max()
    rmax_orig = rmax
    threshold = np.maximum(peak_factor * rmax, threshold)
    # deconvolve
    for i in range(0, maxit):
        modelu = hogbom(mask(residual), psf,
                        gamma=hbgamma, pf=hbpf,
                        maxit=hbmaxit, verbosity=hbverbose)

        model += gamma * modelu

        residual = resid_func(model, dirty, psfo)

        # check stopping criteria
        residual_mfs = np.sum(residual, axis=0)
        rmax = np.abs(residual_mfs).max()
        rms = np.std(residual_mfs)

        if rmax < threshold * rmax_orig:
            print("Success, convergence after %i iterations."
                  "Peak of residual is %f, rms is %f" % (i + 1, rmax, rms),
                  file=log)
            break
        else:
            print("At iteration %i peak of residual is %f, rms is %f" %
                  (i + 1, rmax, rms), file=log)

    return model



# flake8: noqa
from contextlib import ExitStack
from pfb.workers.experimental import cli
from functools import partial
import click
from omegaconf import OmegaConf
import pyscilog
pyscilog.init('pfb')
log = pyscilog.get_logger('BACKWARD')

from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

# create default parameters from schema
defaults = {}
for key in schema.backward["inputs"].keys():
    defaults[key] = schema.backward["inputs"][key]["default"]

@cli.command(context_settings={'show_default': True})
@clickify_parameters(schema.backward)
def backward(**kw):
    '''
    Solves

    argmin_x r(x) + (v - x).H U (v - x) / (2 * gamma)

    If a host address is provided the computation can be distributed
    over imaging band and row. When using a distributed scheduler both
    mem-limit and nthreads is per node and have to be specified.

    When using a local cluster, mem-limit and nthreads refer to the global
    memory and threads available, respectively. By default we will use all
    available resources.

    On a local cluster, the default is to use:

        nworkers = nband
        nthreads-per-worker = 1

    They have to be specified in ~.config/dask/jobqueue.yaml in the
    distributed case.

    if LocalCluster:
        nvthreads = nthreads//(nworkers*nthreads_per_worker)
    else:
        nvthreads = nthreads//nthreads-per-worker

    where nvthreads refers to the number of threads used to scale vertically
    (eg. the number threads given to each gridder instance).

    '''
    defaults.update(kw)
    opts = OmegaConf.create(defaults)
    import time
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    pyscilog.log_to_file(f'backward_{timestamp}.log')

    if opts.nworkers is None:
        if opts.scheduler=='distributed':
            opts.nworkers = opts.nband
        else:
            opts.nworkers = 1

    OmegaConf.set_struct(opts, True)

    with ExitStack() as stack:
        from pfb import set_client
        opts = set_client(opts, stack, log, scheduler=opts.scheduler)

        # TODO - prettier config printing
        print('Input Options:', file=log)
        for key in opts.keys():
            print('     %25s = %s' % (key, opts[key]), file=log)

        return _backward(**opts)

def _backward(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    import numpy as np
    import dask
    import dask.array as da
    import xarray as xr
    from daskms.experimental.zarr import xds_from_zarr, xds_to_zarr
    from pfb.utils.fits import load_fits, set_wcs, save_fits, data_from_header
    from pfb.utils.misc import setup_image_data, init_mask
    from pfb.operators.psi import im2coef, coef2im
    from pfb.operators.hessian import hessian_xds
    from pfb.opt.primal_dual import primal_dual
    from pfb.opt.power_method import power_method
    from pfb.prox.prox_21m import prox_21m
    from pfb.prox.prox_21 import prox_21
    from astropy.io import fits
    import pywt

    basename = f'{opts.output_filename}_{opts.product.upper()}'

    dds_name = f'{basename}{opts.postfix}.dds.zarr'
    mds_name = f'{basename}{opts.postfix}.mds.zarr'

    dds = xds_from_zarr(dds_name, chunks={'row':opts.row_chunk,
                                          'chan': -1})
    # only a single mds (for now)
    mds = xds_from_zarr(mds_name, chunks={'band':1})[0]
    nband = mds.nband
    nx = mds.nx
    ny = mds.ny
    for ds in dds:
        assert ds.nx == nx
        assert ds.ny == ny

    if 'UPDATE' in mds:
        update = mds.UPDATE.values
        assert update.shape == (nband, nx, ny)
    else:
        raise ValueError("No update found in model dataset. "
                         "Use forward worker to populate it. ")

    real_type = dds[0].DIRTY.dtype
    complex_type = np.result_type(real_type, np.complex64)
    if opts.model_name in mds:
        model = mds.get(opts.model_name).values
        assert model.shape == (nband, nx, ny)
        print(f"Initialising model from {opts.model_name} in mds", file=log)
    else:
        print('Initialising model to zeros', file=log)
        model = np.zeros((nband, nx, ny), dtype=real_type)

    data = model + update

    from pfb.utils.misc import init_mask
    mask = init_mask(opts.mask, mds, dds[0].DIRTY.dtype, log)

    # dictionary setup
    print("Setting up dictionary", file=log)
    bases = tuple(opts.bases.split(','))
    nbasis = len(bases)
    iys, sys, ntot, nmax = wavelet_setup(data, bases, opts.nlevels)
    ntot = tuple(ntot)
    psiH = partial(im2coef, bases=bases, ntot=ntot, nmax=nmax,
                   nlevels=opts.nlevels)
    psi = partial(coef2im, bases=bases, ntot=ntot,
                  iy=iys, sy=sys, nx=nx, ny=ny)

    # residual after forward iteration can be useful for setting
    # hyper-parameters
    print("Setting up image space data products", file=log)
    residual, wsum, _, psfhat, mean_beam = setup_image_data(dds, opts,
                                                        'FORWARD_RESIDUAL',
                                                        log=log)
    have_resid = residual.any()
    # we set the alphas used for reweighting using the
    # current clean residuals when available
    if opts.alpha is None:
        alpha = np.ones(nbasis) * 1e-5
        if have_resid:
            resid_comps = psiH(cresid)
            for m in range(nbasis):
                alpha[m] = np.std(resid_comps[:, m])
        else:
            print(f"No residual in dds and alpha was not provided, "
                  f"setting alpha to {alpha[0]}.", file=log)
    else:
        alpha = np.ones(nbasis) * opts.alpha

    if opts.sigma21 is None:
        sigma21 = 1e-4
        if have_resid:
            resid_mfs = np.sum(residual, axis=0)
            sigma21 = np.std(resid_mfs)
        else:
            print("No residual in dds and sigma21 was not provided, "
                  "setting sigma21 to 1e-4.", file=log)
    else:
        sigma21 = opts.sigma21

    hessopts = {}
    hessopts['cell'] = dds[0].cell_rad
    hessopts['wstack'] = opts.wstack
    hessopts['epsilon'] = opts.epsilon
    hessopts['double_accum'] = opts.double_accum
    hessopts['nthreads'] = opts.nvthreads

    if opts.use_psf:
        from pfb.operators.psf import psf_convolve_xds

        nx_psf, ny_psf = dds[0].nx_psf, dds[0].ny_psf
        npad_xl = (nx_psf - nx)//2
        npad_xr = nx_psf - nx - npad_xl
        npad_yl = (ny_psf - ny)//2
        npad_yr = ny_psf - ny - npad_yl
        padding = ((npad_xl, npad_xr), (npad_yl, npad_yr))
        unpad_x = slice(npad_xl, -npad_xr)
        unpad_y = slice(npad_yl, -npad_yr)
        lastsize = ny + np.sum(padding[-1])

        psfopts = {}
        psfopts['padding'] = padding
        psfopts['unpad_x'] = unpad_x
        psfopts['unpad_y'] = unpad_y
        psfopts['lastsize'] = lastsize
        psfopts['nthreads'] = opts.nvthreads

        if opts.mean_ds:
            print("Using mean-ds approximation", file=log)
            from pfb.operators.psf import psf_convolve_cube
            # the PSF is normalised so we don't need to pass wsum
            hess = partial(psf_convolve_cube, psfhat=psfhat,
                           beam=mean_beam * mask[None],
                           psfopts=psfopts, sigmainv=opts.sigmainv,
                           compute=True)
        else:
            from pfb.operators.psf import psf_convolve_xds
            hess = partial(psf_convolve_xds, xds=dds, psfopts=psfopts,
                           wsum=wsum, sigmainv=opts.sigmainv, mask=mask,
                           compute=True)

    else:
        print("Solving for update using vis space approximation", file=log)
        hess = partial(hessian_xds, xds=dds, hessopts=hessopts,
                        wsum=wsum, sigmainv=opts.sigmainv, mask=mask,
                        compute=True)

    if opts.hessnorm is None:
        print("Finding spectral norm of Hessian approximation", file=log)
        hessnorm, _ = power_method(hess, (nband, nx, ny), tol=opts.pm_tol,
                         maxit=opts.pm_maxit, verbosity=opts.pm_verbose,
                         report_freq=opts.pm_report_freq)
    else:
        hessnorm = opts.hessnorm

    if 'DUAL' in mds:
        dual = mds.DUAL.values
        assert dual.shape == (nband, nbasis, nmax)
    else:
        dual = np.zeros((nband, nbasis, nmax), dtype=model.dtype)

    if 'WEIGHT' in mds:
        weight = mds.WEIGHT.values
        assert weight.shape == (nbasis, nmax)
    else:
        weight = np.ones((nbasis, nmax), dtype=model.dtype)

    modelp = model.copy()
    print("Solving for model", file=log)
    for i in range(opts.niter):
        # prox = partial(prox_21m, sigma=opts.sigma21, weight=weight, axis=0)
        model, dual = primal_dual(hess, data, model, dual, sigma21,
                                  psi, psiH, weight, hessnorm, prox_21,
                                  nu=nbasis, positivity=opts.positivity,
                                  tol=opts.pd_tol, maxit=opts.pd_maxit,
                                  verbosity=opts.pd_verbose,
                                  report_freq=opts.pd_report_freq)

        # reweight
        l2_norm = np.linalg.norm(psiH(model), axis=0)
        for m in range(nbasis):
            # if adapt_sig21:
            #     _, sigmas[m] = expon.fit(l2_norm[m], floc=0.0)
            #     print('basis %i, sigma %f'%sigmas[m], file=log)

            weight[m] = alpha[m]/(alpha[m] + l2_norm[m])

    if opts.niter==0:
        model = data

    print("Saving results", file=log)
    mask = np.any(model, axis=0).astype(bool)
    mask = da.from_array(mask, chunks=(-1, -1))
    model = da.from_array(model, chunks=(1, -1, -1), name=False)
    modelp = da.from_array(modelp, chunks=(1, -1, -1), name=False)
    dual = da.from_array(dual, chunks=(1, -1, -1), name=False)
    weight = da.from_array(weight, chunks=(-1, -1), name=False)

    mds = mds.assign(**{
                     'MASK': (('x', 'y'), mask),
                     'MODEL': (('band', 'x', 'y'), model),
                     'MODELP': (('band', 'x', 'y'), modelp),
                     'DUAL': (('band', 'basis', 'coef'), dual),
                     'WEIGHT': (('basis', 'coef'), weight)})

    dask.compute(xds_to_zarr(mds, mds_name, columns='ALL'))

    if opts.do_residual:
        print("Computing residual", file=log)
        # compute apparent residual per dataset
        from pfb.operators.hessian import hessian
        # Required because of https://github.com/ska-sa/dask-ms/issues/171
        ddsw = xds_from_zarr(dds_name, columns='DIRTY')
        writes = []
        for ds, dsw in zip(dds, ddsw):
            dirty = ds.DIRTY.data
            wgt = ds.WEIGHT.data
            uvw = ds.UVW.data
            freq = ds.FREQ.data
            beam = ds.BEAM.data
            vis_mask = ds.MASK.data
            b = ds.bandid
            # we only want to apply the beam once here
            residual = (dirty -
                        hessian(beam * model[b], uvw, wgt, vis_mask, freq,
                                None, hessopts))
            dsw = dsw.assign(**{'RESIDUAL': (('x', 'y'), residual)})
            writes.append(dsw)

        dask.compute(xds_to_zarr(writes, dds_name, columns='RESIDUAL'))

    if opts.fits_mfs or opts.fits_cubes:
        print("Writing fits files", file=log)
        dds = xds_from_zarr(dds_name)

        # construct a header from xds attrs
        ra = mds.ra
        dec = mds.dec
        radec = [ra, dec]

        cell_rad = mds.cell_rad
        cell_deg = np.rad2deg(cell_rad)

        freq_out = mds.band.values
        ref_freq = np.mean(freq_out)
        hdr_mfs = set_wcs(cell_deg, cell_deg, nx, ny, radec, ref_freq)

        model_mfs = np.mean(model, axis=0)
        save_fits(f'{basename}_model_mfs.fits', model_mfs, hdr_mfs)

        if opts.do_residual:
            residual = np.zeros((nband, nx, ny), dtype=dds[0].DIRTY.dtype)
            wsums = np.zeros(nband)
            for ds in dds:
                b = ds.bandid
                wsums[b] += ds.WSUM.values[0]
                residual[b] += ds.RESIDUAL.values
            wsum = np.sum(wsums)
            residual /= wsum
            residual_mfs = np.sum(residual, axis=0)
            save_fits(f'{basename}_residual_mfs.fits', residual_mfs, hdr_mfs)

        if opts.fits_cubes:
            # need residual in Jy/beam
            wsums = np.amax(psf, axes=(1,2))
            hdr = set_wcs(cell_deg, cell_deg, nx, ny, radec, freq_out)
            save_fits(f'{basename}_model.fits', model, hdr)
            if opts.do_residual:
                fmask = wsums > 0
                residual[fmask] /= wsums[fmask, None, None]
                save_fits(f'{basename}_residual.fits',
                        residual, hdr)

    print("All done here.", file=log)



import numba
from numba.core import types, cgutils, typing
from numba.extending import intrinsic
from numba.np.arrayobj import make_view, _change_dtype
from numba.core.imputils import impl_ret_borrowed


@intrinsic
def force_type_contiguity(typingctx, array):
    return_type = array.copy(layout="C")
    sig = return_type(array)

    def codegen(context, builder, signature, args):
        def check(array):
            if not array.flags.c_contiguous:
                raise ValueError("Attempted to force type contiguity "
                                 "on an array whose layout is "
                                 "non-contiguous")

        check_sig = typing.signature(types.none, signature.args[0])
        context.compile_internal(builder, check, check_sig, args)

        return impl_ret_borrowed(context, builder, return_type, args[0])

    return sig, codegen


@intrinsic
def slice_axis(typingctx, array, index, axis, extent):
    if not isinstance(array, types.Array):
        raise TypeError("array is not an Array")

    if "C" not in array.layout:
        raise TypeError("array must be C contiguous")

    if (not isinstance(index, types.UniTuple) or
            not isinstance(index.dtype, types.Integer)):
        raise TypeError("index is not an Homogenous Tuple of Integers")

    if not isinstance(axis, types.Integer):
        raise TypeError("axis is not an Integer")

    have_extent = not cgutils.is_nonelike(extent)

    if have_extent and not isinstance(extent, types.Integer):
        raise TypeError("extent must be an integer or None")

    if len(index) != array.ndim:
        raise TypeError("array.ndim != len(index")

    # Return a single array, not necessarily contiguous
    return_type = array.copy(ndim=1, layout="A")

    sig = return_type(array, index, axis, extent)

    def codegen(context, builder, signature, args):
        array_type, idx_type, axis_type, extent_type = signature.args
        array, idx, axis, extent = args
        array = context.make_array(array_type)(context, builder, array)

        zero = context.get_constant(types.intp, 0)
        llvm_intp_t = context.get_value_type(types.intp)
        ndim = array_type.ndim

        view_shape = cgutils.alloca_once(builder, llvm_intp_t)
        view_stride = cgutils.alloca_once(builder, llvm_intp_t)

        # Final array indexes. We only know the slicing index at runtime
        # so we need to recreate idx but with zero at the slicing axis
        indices = cgutils.alloca_once(
            builder, llvm_intp_t, size=array_type.ndim)

        for ax in range(array_type.ndim):
            llvm_ax = context.get_constant(types.intp, ax)
            predicate = builder.icmp_unsigned("!=", llvm_ax, axis)

            with builder.if_else(predicate) as (not_equal, equal):
                with not_equal:
                    # If this is not the slicing axis,
                    # use the appropriate tuple index
                    value = builder.extract_value(idx, ax)
                    builder.store(value, builder.gep(indices, [llvm_ax]))

                with equal:
                    # If this is the slicing axis,
                    # store zero as the index.
                    # Also record the stride and shape
                    builder.store(zero, builder.gep(indices, [llvm_ax]))
                    size = builder.extract_value(array.shape, ax)
                    stride = builder.extract_value(array.strides, ax)

                    if have_extent:
                        ext_predicate = builder.icmp_signed(">=", extent, size)
                        size = builder.select(ext_predicate, size, extent)

                    builder.store(size, view_shape)
                    builder.store(stride, view_stride)

        # Build a python list from indices
        tmp_indices = []

        for i in range(ndim):
            i = context.get_constant(types.intp, i)
            tmp_indices.append(builder.load(builder.gep(indices, [i])))

        # Get the data pointer obtained from indexing the array
        dataptr = cgutils.get_item_pointer(context, builder,
                                           array_type, array,
                                           tmp_indices,
                                           wraparound=True,
                                           boundscheck=True)

        # Set up the shape and stride. There'll only be one
        # dimension, corresponding to the axis along which we slice
        view_shapes = [builder.load(view_shape)]
        view_strides = [builder.load(view_stride)]

        # Make a view with the data pointer, shapes and strides
        retary = make_view(context, builder,
                           array_type, array, return_type,
                           dataptr, view_shapes, view_strides)

        result = retary._getvalue()
        return impl_ret_borrowed(context, builder, return_type, result)

    return sig, codegen


@intrinsic
def not_optional(typingctx, value):
    # Copied from numba/typed/typedobjectutils.py::_nonoptional
    if not isinstance(value, types.Optional):
        raise TypeError(f"value should be an Optional. {value}")

    def codegen(context, builder, sig, args):
        context.nrt.incref(builder, sig.return_type, args[0])
        return args[0]

    casted = value.type
    sig = casted(casted)
    return sig, codegen



from setuptools import setup, find_packages
import pfb

with open("README.rst", "r") as fh:
    long_description = fh.read()

requirements = [
                'numpy',
                'matplotlib',
                'ipython',
                'scikit-image',
                'dask[distributed]',
                'PyWavelets',
                'katbeam',
                'pytest >= 6.2.2',
                'numexpr',
                'pyscilog >= 0.1.2',
                'Click',
                'omegaconf',
                "codex-africanus[dask, scipy, astropy, python-casacore, ducc0]",
                "dask-ms[xarray, zarr, s3]",
                "stimela==2.0rc6",
                "QuartiCal",
                "sympy",

            ]


setup(
     name='pfb-clean',
     version=pfb.__version__,
     author="Landman Bester",
     author_email="lbester@sarao.ac.za",
     description="Pre-conditioned forward-backward CLEAN algorithm",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://github.com/ratt-ru/pfb-clean",
     packages=find_packages(),
     include_package_data=True,
     zip_safe=False,
     python_requires='>=3.8',
     install_requires=requirements,
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: POSIX :: Linux",
         "Topic :: Scientific/Engineering :: Astronomy",
     ],
     entry_points={'console_scripts':[
        'pfb = pfb.workers.main:cli'
        ]
     }


     ,
 )



# flake8: noqa
from contextlib import ExitStack
from pfb.workers.main import cli
import click
from omegaconf import OmegaConf
import pyscilog
pyscilog.init('pfb')
log = pyscilog.get_logger('RESTORE')

from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

# create default parameters from schema
defaults = {}
for key in schema.restore["inputs"].keys():
    defaults[key] = schema.restore["inputs"][key]["default"]

@cli.command(context_settings={'show_default': True})
@clickify_parameters(schema.restore)
def restore(**kw):
    '''
    Create fits image data products (eg. restored images).
    '''
    defaults.update(kw)
    opts = OmegaConf.create(defaults)
    pyscilog.log_to_file(f'{opts.output_filename}_{opts.product}{opts.postfix}.log')
    if opts.nworkers is None:
        opts.nworkers = opts.nband

    OmegaConf.set_struct(opts, True)

    with ExitStack() as stack:
        # numpy imports have to happen after this step
        from pfb import set_client
        set_client(opts, stack, log, scheduler=opts.scheduler)

        # TODO - prettier config printing
        print('Input Options:', file=log)
        for key in opts.keys():
            print('     %25s = %s' % (key, opts[key]), file=log)

        return _restore(**opts)

def _restore(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)


    import numpy as np
    from daskms.experimental.zarr import xds_from_zarr
    from pfb.utils.fits import (save_fits, add_beampars, set_wcs,
                                dds2fits, dds2fits_mfs)
    from pfb.utils.misc import Gaussian2D, fitcleanbeam, convolve2gaussres, dds2cubes

    basename = f'{opts.output_filename}_{opts.product.upper()}'
    dds_name = f'{basename}_{opts.postfix}.dds'

    dds = xds_from_zarr(dds_name)
    nband = opts.nband
    nx = dds[0].x.size
    ny = dds[0].y.size
    cell_rad = dds[0].cell_rad
    cell_deg = np.rad2deg(cell_rad)
    freq = []
    for ds in dds:
        freq.append(ds.freq_out)
        assert ds.x.size == nx
        assert ds.y.size == ny
    freq = np.unique(np.array(freq))
    assert freq.size == opts.nband

    # init fits headers
    radec = (dds[0].ra, dds[0].dec)
    ref_freq = np.mean(freq)
    hdr_mfs = set_wcs(cell_deg, cell_deg, nx, ny, radec, ref_freq)
    hdr = set_wcs(cell_deg, cell_deg, nx, ny, radec, freq)


    dirty, model, residual, psf, _, _, wsums, _ = dds2cubes(dds,
                                                            nband,
                                                            apparent=True)
    wsum = np.sum(wsums)
    output_type = dirty.dtype
    fmask = wsums > 0

    if residual is None:
        print('Warning, no residual in dds. '
              'Using dirty as residual.', file=log)
        residual = dirty.copy()
    residual_mfs = np.sum(residual, axis=0)
    residual[fmask] /= wsums[fmask, None, None]/wsum

    if not model.any():
        print("Warning - model is empty", file=log)
    model_mfs = np.mean(model, axis=0)

    if psf is not None:
        nx_psf = dds[0].x_psf.size
        ny_psf = dds[0].y_psf.size
        psf_mfs = np.sum(psf, axis=0)
        psf[fmask] /= wsums[fmask, None, None]/wsum
        # sanity check
        assert (psf_mfs.max() - 1.0) < 2e-7
        assert ((np.amax(psf, axis=(1,2)) - 1.0) < 2e-7).all()

        # fit restoring psf
        GaussPar = fitcleanbeam(psf_mfs[None], level=0.5, pixsize=1.0)
        cpsf_mfs = np.zeros(residual_mfs.shape, dtype=output_type)
        lpsf = -(nx//2) + np.arange(nx)
        mpsf = -(ny//2) + np.arange(ny)
        xx, yy = np.meshgrid(lpsf, mpsf, indexing='ij')
        cpsf_mfs = Gaussian2D(xx, yy, GaussPar[0], normalise=False)
        image_mfs = convolve2gaussres(model_mfs[None], xx, yy,
                                    GaussPar[0], opts.nvthreads,
                                    norm_kernel=False)[0]  # peak of kernel set to unity
        image_mfs += residual_mfs
        # convert pixel units to deg
        GaussPar[0][0] *= cell_deg
        GaussPar[0][1] *= cell_deg
        hdr_mfs = add_beampars(hdr_mfs, GaussPar)

        if any([i.isupper() for i in opts.outputs]):
            GaussPars = fitcleanbeam(psf, level=0.5, pixsize=1.0)  # pixel units
            cpsf = np.zeros(residual.shape, dtype=output_type)
            for v in range(opts.nband):
                cpsf[v] = Gaussian2D(xx, yy, GaussPars[v], normalise=False)

            image = np.zeros_like(model)
            for b in range(nband):
                image[b:b+1] = convolve2gaussres(model[b:b+1], xx, yy,
                                                GaussPars[b], opts.nvthreads,
                                                norm_kernel=False)  # peak of kernel set to unity
                image[b] += residual[b]

            for i, gp in enumerate(GaussPars):
                GaussPars[i] = [gp[0]*cell_deg, gp[1]*cell_deg, gp[2]]

            hdr = add_beampars(hdr, GaussPar, GaussPars)
    else:
        print('Warning, no psf in dds. '
              'Unable to add resolution info or make restored image. ',
              file=log)

    if 'm' in opts.outputs:
        save_fits(model_mfs,
                  f'{basename}_{opts.postfix}.model_mfs.fits',
                  hdr_mfs,
                  overwrite=opts.overwrite)

    if 'M' in opts.outputs:
        save_fits(model,
                  f'{basename}_{opts.postfix}.model.fits',
                  hdr,
                  overwrite=opts.overwrite)

    if 'r' in opts.outputs:
        save_fits(residual_mfs,
                  f'{basename}_{opts.postfix}.residual_mfs.fits',
                  hdr_mfs,
                  overwrite=opts.overwrite)

    if 'R' in opts.outputs:
        save_fits(residual,
                  f'{basename}_{opts.postfix}.residual.fits',
                  hdr,
                  overwrite=opts.overwrite)

    if 'd' in opts.outputs:
        dirty_mfs = np.sum(dirty, axis=0)
        save_fits(dirty_mfs,
                  f'{basename}_{opts.postfix}.dirty_mfs.fits',
                  hdr_mfs,
                  overwrite=opts.overwrite)

    if 'D' in opts.outputs:
        dirty[fmask] /= wsums[fmask, None, None]/wsum
        save_fits(dirty,
                  f'{basename}_{opts.postfix}.dirty.fits',
                  hdr,
                  overwrite=opts.overwrite)

    if 'i' in opts.outputs and psf is not None:
        save_fits(image_mfs,
                  f'{basename}_{opts.postfix}.image_mfs.fits',
                  hdr_mfs,
                  overwrite=opts.overwrite)

    if 'I' in opts.outputs and psf is not None:
        save_fits(image,
                  f'{basename}_{opts.postfix}.image.fits',
                  hdr,
                  overwrite=opts.overwrite)

    if 'c' in opts.outputs and psf is not None:
        save_fits(cpsf_mfs,
                  f'{basename}_{opts.postfix}.cpsf_mfs.fits',
                  hdr_mfs,
                  overwrite=opts.overwrite)

    if 'C' in opts.outputs and psf is not None:
        save_fits(cpsf,
                  f'{basename}_{opts.postfix}.cpsf.fits',
                  hdr,
                  overwrite=opts.overwrite)

    if opts.scheduler=='distributed':
        from distributed import get_client
        client = get_client()
        client.close()

    print("All done here", file=log)



import numpy as np
import dask
import dask.array as da
from daskms.optimisation import inlined_array
from ducc0.wgridder.experimental import vis2dirty, dirty2vis
from ducc0.misc import make_noncritical
from uuid import uuid4
from pfb.operators.psf import (psf_convolve_slice,
                               psf_convolve_cube)


def hessian_xds(x, xds, hessopts, wsum, sigmainv, mask,
                compute=True, use_beam=True):
    '''
    Vis space Hessian reduction over dataset.
    Hessian will be applied to x
    '''
    if not isinstance(x, da.Array):
        x = da.from_array(x, chunks=(1, -1, -1),
                          name="x-" + uuid4().hex)

    if not isinstance(mask, da.Array):
        mask = da.from_array(mask, chunks=(-1, -1),
                             name="mask-" + uuid4().hex)

    assert mask.ndim == 2

    nband, nx, ny = x.shape

    # LB - what is the point of specifying name?
    convims = [da.zeros((nx, ny),
               chunks=(-1, -1), name="zeros-" + uuid4().hex)
               for _ in range(nband)]

    for ds in xds:
        wgt = ds.WEIGHT.data
        vis_mask = ds.MASK.data
        uvw = ds.UVW.data
        freq = ds.FREQ.data
        b = ds.bandid
        if use_beam:
            beam = ds.BEAM.data * mask
        else:
            # TODO - separate implementation without
            # unnecessary beam application
            beam = mask

        convim = hessian(x[b], uvw, wgt, vis_mask, freq, beam, hessopts)

        # convim = inlined_array(convim, uvw)

        convims[b] += convim

    convim = da.stack(convims)/wsum

    if sigmainv:
        convim += x * sigmainv**2

    if compute:
        return convim.compute()
    else:
        return convim


def _hessian_impl(x, uvw, weight, vis_mask, freq, beam,
                  x0=0.0,
                  y0=0.0,
                  cell=None,
                  wstack=None,
                  epsilon=None,
                  double_accum=None,
                  nthreads=None):
    if not x.any():
        return np.zeros_like(x)
    nx, ny = x.shape
    mvis = dirty2vis(uvw=uvw,
                    freq=freq,
                    mask=vis_mask,
                    dirty=x if beam is None else x * beam,
                    pixsize_x=cell,
                    pixsize_y=cell,
                    center_x=x0,
                    center_y=y0,
                    epsilon=epsilon,
                    nthreads=nthreads,
                    do_wgridding=wstack,
                    divide_by_n=False)

    convim = vis2dirty(uvw=uvw,
                      freq=freq,
                      vis=mvis,
                      wgt=weight,
                      mask=vis_mask,
                      npix_x=nx,
                      npix_y=ny,
                      pixsize_x=cell,
                      pixsize_y=cell,
                      center_x=x0,
                      center_y=y0,
                      epsilon=epsilon,
                      nthreads=nthreads,
                      do_wgridding=wstack,
                      double_precision_accumulation=double_accum,
                      divide_by_n=False)

    if beam is not None:
        convim *= beam

    return convim


def _hessian(x, uvw, weight, vis_mask, freq, beam, hessopts):
    return _hessian_impl(x, uvw[0][0], weight[0][0], vis_mask[0][0], freq[0],
                         beam, **hessopts)

def hessian(x, uvw, weight, vis_mask, freq, beam, hessopts):
    if beam is None:
        bout = None
    else:
        bout = ('nx', 'ny')
    return da.blockwise(_hessian, ('nx', 'ny'),
                        x, ('nx', 'ny'),
                        uvw, ('row', 'three'),
                        weight, ('row', 'chan'),
                        vis_mask, ('row', 'chan'),
                        freq, ('chan',),
                        beam, bout,
                        hessopts, None,
                        dtype=x.dtype)


def _hessian_psf_slice(
                    xpad,  # preallocated array to store padded image
                    xhat,  # preallocated array to store FTd image
                    xout,  # preallocated array to store output image
                    psfhat,
                    beam,
                    lastsize,
                    x,     # input image, not overwritten
                    nthreads=1,
                    sigmainv=1,
                    wsum=None):
    """
    Tikhonov regularised Hessian approx
    """
    if beam is not None:
        psf_convolve_slice(xpad, xhat, xout,
                           psfhat, lastsize, x*beam,
                           nthreads=nthreads)
    else:
        psf_convolve_slice(xpad, xhat, xout,
                           psfhat, lastsize, x,
                           nthreads=nthreads)

    if beam is not None:
        xout *= beam

    if wsum is not None:
        xout /= wsum

    return xout + x * sigmainv

from pfb.operators.hessian import _hessian_impl
class hessian_psf_slice(object):
    def __init__(self, ds, nbasis, nmax, nthreads, sigmainv, cell, wstack, epsilon, double_accum):
        self.nthreads = nthreads
        self.sigmainv = sigmainv
        self.cell = cell
        self.wstack = wstack
        self.epsilon = epsilon
        self.double_accum = double_accum
        self.lastsize = ds.PSF.shape[-1]
        self.bandid = ds.bandid
        tmp = np.require(ds.DIRTY.values,
                         dtype=ds.DIRTY.dtype,
                         requirements='CAW')
        self.dirty = make_noncritical(tmp)
        tmp = np.require(ds.PSFHAT.values,
                         dtype=ds.PSFHAT.dtype,
                         requirements='CAW')
        self.psfhat = make_noncritical(tmp)
        tmp = np.require(ds.PSF.values,
                         dtype=ds.PSF.dtype,
                         requirements='CAW')
        self.psf = make_noncritical(tmp)
        tmp = np.require(ds.BEAM.values,
                         dtype=ds.BEAM.dtype,
                         requirements='CAW')
        self.beam = make_noncritical(tmp)
        self.wsumb = ds.WSUM.values[0]
        if 'MODEL' in ds:
            tmp = np.require(ds.MODEL.values,
                             dtype=ds.MODEL.dtype,
                             requirements='CAW')
        else:
            tmp = np.zeros_like(self.dirty)
        self.model = make_noncritical(tmp)
        if 'DUAL' in ds:
            tmp = np.require(ds.DUAL.values,
                             dtype=ds.DUAL.dtype,
                             requirements='CAW')
            assert tmp.shape == (nbasis, nmax)
        else:
            tmp = np.zeros((nbasis, nmax), dtype=self.dirty.dtype)
        self.dual = make_noncritical(tmp)
        if 'RESIDUAL' in ds:
            tmp = np.require(ds.RESIDUAL.values,
                             dtype=ds.RESIDUAL.dtype,
                             requirements='CAW')
        else:
            tmp = self.dirty.copy()
        self.residual = make_noncritical(tmp)

        self.uvw = ds.UVW.values
        self.wgt = ds.WEIGHT.values
        self.vmask = ds.VIS_MASK.values
        self.freq = ds.FREQ.values

        # pre-allocate tmp arrays
        tmp = np.empty(self.dirty.shape, dtype=self.dirty.dtype, order='C')
        self.xout = make_noncritical(tmp)
        tmp = np.empty(self.psfhat.shape, dtype=self.psfhat.dtype, order='C')
        self.xhat = make_noncritical(tmp)
        tmp = np.empty(self.psf.shape, dtype=self.psf.dtype, order='C')
        self.xpad = make_noncritical(tmp)

    def __call__(self, x):
        return _hessian_psf_slice(self.xpad,
                                  self.xhat,
                                  self.xout,
                                  self.psfhat,
                                  self.beam,
                                  self.lastsize,
                                  x,
                                  nthreads=self.nthreads,
                                  sigmainv=self.sigmainv,
                                  wsum=self.wsum)

    def compute_residual(self, x):
        return self.dirty - _hessian_impl(self.beam * x,
                                        self.uvw,
                                        self.wgt,
                                        self.vmask,
                                        self.freq,
                                        None,
                                        cell=self.cell,
                                        wstack=self.wstack,
                                        epsilon=self.epsilon,
                                        double_accum=self.double_accum,
                                        nthreads=self.nthreads)


    def set_wsum(self, wsum):
        self.wsum = wsum


def hessian_psf_cube(
                    xpad,  # preallocated array to store padded image
                    xhat,  # preallocated array to store FTd image
                    xout,  # preallocated array to store output image
                    beam,
                    psfhat,
                    lastsize,
                    x,     # input image, not overwritten
                    nthreads=1,
                    sigmainv=1,
                    wsum=None):
    """
    Tikhonov regularised Hessian approx
    """
    if beam is not None:
        psf_convolve_cube(xpad, xhat, xout, psfhat, lastsize, x*beam,
                          nthreads=nthreads)
    else:
        psf_convolve_cube(xpad, xhat, xout, psfhat, lastsize, x,
                          nthreads=nthreads)

    if beam is not None:
        xout *= beam

    if wsum is not None:
        xout /= wsum

    return xout + x * sigmainv


def hess_vis(xds,
             dds,
             xout,
             x,
             sigmainv=1.0,
             wstack=True,
             nthreads=1,
             epsilon=1e-7,
             divide_by_n=False):
    for ds in xds:
        b = ds.bandid
        t = ds.timeid
        vis_mask = ds.MASK.values
        if np.all(vis_mask == 0):
            continue

        # accumulate model vis for this band and time
        mvis = np.zeros(ds.VIS.data.shape, dtype=ds.VIS.dtype)
        for field in dds.keys():

            x0 = dds[field][f't{t}b{b}']['x0']
            y0 = dds[field][f't{t}b{b}']['y0']
            cell = dds[field][f't{t}b{b}']['cell']
            nx = dds[field][f't{t}b{b}']['nx']
            ny = dds[field][f't{t}b{b}']['ny']

            mvis += dirty2vis(uvw=ds.UVW.values,
                              freq=ds.FREQ.values,
                              dirty=x[field][f't{t}b{b}'],
                              pixsize_x=cell,
                              pixsize_y=cell,
                              center_x=x0,
                              center_y=y0,
                              epsilon=epsilon,
                              do_wgridding=wstack,
                              nthreads=nthreads,
                              divide_by_n=divide_by_n)

        # project to image space
        for field in dds.keys():
            x0 = dds[field][f't{t}b{b}']['x0']
            y0 = dds[field][f't{t}b{b}']['y0']
            cell = dds[field][f't{t}b{b}']['cell']
            nx = dds[field][f't{t}b{b}']['nx']
            ny = dds[field][f't{t}b{b}']['ny']
            xout[field][f't{t}b{b}'] = vis2dirty(uvw=ds.UVW.values,
                                                freq=ds.FREQ.values,
                                                vis=mvis,
                                                wgt=ds.WEIGHT.values,
                                                npix_x=nx,
                                                npix_y=ny,
                                                pixsize_x=cell,
                                                pixsize_y=cell,
                                                center_x=x0,
                                                center_y=y0,
                                                epsilon=epsilon,
                                                do_wgridding=wstack,
                                                nthreads=nthreads,
                                                divide_by_n=divide_by_n)
            xout[field][f't{t}b{b}'] += sigmainv * x[field][f't{t}b{b}']
    return xout



import numpy as np
from functools import partial
from katbeam import JimBeam
import dask.array as da

def Id_beam(l, m):
    return np.ones(l.shape, dtype=float)

def _interp_beam_impl(freq, nx, ny, cell_deg, btype):
    '''
    A function that returns an object array containing a function
    returning beam values given (l,m) coordinates at a single frequency.
    Frequency mapped to imaging band extenally. Result is meant to be
    passed into eval_beam below.
    '''
    l = (-(nx//2) + np.arange(nx)) * cell_deg
    m = (-(ny//2) + np.arange(ny)) * cell_deg
    ll, mm = np.meshgrid(l, m, indexing='ij')

    if btype is None:
        beam = Id_beam
    else:
        btype = btype.lower()
        btype = btype.replace('-', '_')
        if btype in ["kbl", "kb_l", "katbeam_l"]:
            # katbeam L band
            beam = partial(JimBeam('MKAT-AA-L-JIM-2020').I, freqMHz=freq)
        elif btype in ["kbuhf", "kb_uhf", "katbeam_uhf"]:
            # katbeam L band
            beam = partial(JimBeam('MKAT-AA-UHF-JIM-2020').I, freqMHz=freq)
        else:
            raise ValueError(f"Unknown beam model {btype}")
    return np.array([beam], dtype=object)


def interp_beam(freq, nx, ny, cell_deg, btype):
    '''
    Blockwise wrapper that returns an object array containing a function
    returning beam values given (l,m) coordinates at a single frequency.
    Frequency mapped to imaging band extenally. Result is meant to be
    passed into eval_beam below.
    '''
    return da.blockwise(_interp_beam_impl, '1',
                        freq, None,
                        nx, None,
                        ny, None,
                        cell_deg, None,
                        btype, None,
                        new_axes={'1': 1},
                        dtype=object)

def _eval_beam_impl(beam_object_array, l, m):
    return beam_object_array[0](l, m)

def _eval_beam(beam_object_array, l, m):
    return _eval_beam_impl(beam_object_array[0], l, m)

def eval_beam(beam_object_array, l, m):
    if l.ndim == 2:
        lout = ('nx', 'ny')
        mout = ('nx', 'ny')
    else:
        lout = ('nx',)
        mout = ('ny',)
    return da.blockwise(_eval_beam, ('nx', 'ny'),
                        beam_object_array, ('1',),
                        l, lout,
                        m, mout,
                        dtype=float)
